#include "ros/ros.h"
#include <sensor_msgs/PointCloud2.h>
#include <pcl_conversions/pcl_conversions.h>
#include <geometry_msgs/Pose.h>
#include <eigen_conversions/eigen_msg.h> //tf::poseEigenToMsg


#include <pcl/point_types.h>
#include <pcl/point_cloud.h>
#include <pcl/common/time.h>
#include <pcl/console/print.h>
#include <pcl/io/pcd_io.h>
#include <pcl/visualization/pcl_visualizer.h>
#include <pcl/search/kdtree.h>
#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/filters/extract_indices.h>
#include <pcl/filters/project_inliers.h>

#include "../include/wm_segmentation.h"
#include "../include/find_target.h"

#include "vision_msgs/SegCloud.h"

std::string object_file = "/home/utente/baxter_ws/src/vision/wm_cloud_segmentation/pcds/model_wm_baxter.pcd";


// Types
typedef pcl::PointNormal PointNT;
typedef pcl::PointXYZI PointT;
typedef pcl::PointCloud<PointNT> PointCloudT;
typedef pcl::PointCloud<PointT> PointCloudIntT;


class Segment{

    std::string topic_cloud_, topic_srv_, object_file_;
    float wm_seg_threshold_;

    ros::NodeHandle n_;
    ros::ServiceServer get_srv_;

public:
    Segment(ros::NodeHandle* n);

    void maincall();

    bool serviceCallback(   vision_msgs::SegCloud::Request &req,
                            vision_msgs::SegCloud::Response &res);
};

Segment::Segment(ros::NodeHandle* n):n_(*n) 
{
    std::cout<<"Constructor started"<<std::endl;

    n_.getParam("wm_seg/topic_cloud", topic_cloud_);
    n_.getParam("wm_seg/topic_srv", topic_srv_);
    n_.getParam("wm_seg/object_file", object_file_);
    n_.getParam("wm_seg/seg_threshold", wm_seg_threshold_);

    get_srv_ = n_.advertiseService(topic_srv_, &Segment::serviceCallback, this);

    std::cout<<"Constructor finished"<<std::endl;
}

bool Segment::serviceCallback(vision_msgs::SegCloud::Request &req, vision_msgs::SegCloud::Response &res){
    // -----------------------
    // Point clouds
    PointCloudT::Ptr object(new PointCloudT);
    PointCloudT::Ptr object_result(new PointCloudT);
    PointCloudT::Ptr scene(new PointCloudT);
    pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_difference(new pcl::PointCloud<pcl::PointXYZRGB>);


    // Load object 
    pcl::console::print_highlight("Loading OBJECT cloud...\n");
    if (pcl::io::loadPCDFile<pcl::PointNormal>(object_file_, *object) < 0)
    {
        pcl::console::print_error("Error loading object file!\n");
        return (1);
    }
    pcl::console::print_highlight("done\n");

    pcl::console::print_highlight("Loading SCENE cloud...\n");
    // GET SCENE CLOUD --------------
    sensor_msgs::PointCloud2ConstPtr sharedMsg;
    sensor_msgs::PointCloud2 msg;
    sharedMsg = ros::topic::waitForMessage<sensor_msgs::PointCloud2>(topic_cloud_, n_);
    if (sharedMsg != NULL)
        msg = *sharedMsg;

    pcl::PCLPointCloud2 pcl_pc2;
    pcl_conversions::toPCL(*sharedMsg, pcl_pc2);
    PointCloudT::Ptr cloud(new PointCloudT);
    pcl::fromPCLPointCloud2(pcl_pc2, *scene);
    pcl::console::print_highlight("done\n");

    std::cout << "scene cloud: " << scene->size() << std::endl;
    std::cout << "object cloud: " << object->size() << std::endl;

    pcl::console::print_highlight("Finding Washing Machine...\n");
    FindTarget ft;
    ft.object = object;
    ft.scene = scene;
    ft.apply_icp = false;
    ft.compute();

    object_result = ft.object_icp;
    std::cout << "object_result cloud: " << object_result->size() << std::endl;

    pcl::console::print_highlight("Computing Difference Map...\n");
    WMSegmentation seg;
    seg.setObjectCloud(object_result);
    seg.setSceneCloud(scene);
    seg.setSegmentationThreshold(wm_seg_threshold_);
    pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_entropy(new pcl::PointCloud<pcl::PointXYZRGB>);
    seg.compute(cloud_entropy);

    std::cout << "cloud_entropy size: " << cloud_entropy->size() << std::endl;

    pcl::ModelCoefficients::Ptr plane = seg.getPlane();
    std::cout << "PLANE: " << plane->values[0] << " " << plane->values[1]
              << " " << plane->values[2] << " " << plane->values[3] << std::endl;

    
    // TRANSFORM TO POINTCLOUD2 TYPE
    sensor_msgs::PointCloud2 cloud_output;
    pcl::toROSMsg(*cloud_entropy, cloud_output);

    // TRANSFROM PLANE TO FLOAT64 VECTOR
    std::vector<double> plane_output = {plane->values[0], plane->values[1], plane->values[2], plane->values[3]};

    res.cloud_seg = cloud_output;
    res.plane = plane_output;

    return true;


}


// Align a rigid object to a scene with clutter and occlusions
int main(int argc, char **argv)
{
    ros::init(argc, argv, "wm_segmentation_node");
    ros::NodeHandle n;

    Segment s(&n);

    ROS_INFO("Service Ready");
    ros::spin();
 
    return (0);
}
