#include "../include/wm_segmentation.h"

void WMSegmentation::getPointPlaneDistanceCloud(PointCloudT::Ptr &obj, PointCloudT::Ptr &scene, PointCloudT::Ptr scene_proj,
                                                PointCloudIntT::Ptr &cloud_out, pcl::ModelCoefficients::Ptr &plane)
{
    float threshold = 0.05;

    Eigen::Vector3f x0, y0;
    Eigen::Vector3f n(plane->values[0], plane->values[1], plane->values[2]);
    float p = plane->values[3];
    float distance, distance_obj;

    pcl::search::KdTree<PointNT> kdtree;
    PointNT searchPoint;
    kdtree.setInputCloud(scene_proj);

    int K = 1;
    std::vector<int> indices(K);
    std::vector<float> distances(K);



    for (int i = 0; i < obj->size(); i++)
    {

        searchPoint = obj->points[i];

        if (kdtree.nearestKSearch(searchPoint, K, indices, distances) > 0)
        {
            

            if ((sqrt(distances[0]) < threshold) &&
                (!std::isnan(scene->points[indices[0]].x)) &&
                (!std::isnan(scene->points[indices[0]].y)) && (!std::isnan(scene->points[indices[0]].z))) 
            {

                // computation for scene cloud
                x0 = scene->points[indices[0]].getVector3fMap();
                distance = n.dot(x0) + p;

                // computation for obj cloud
                y0 = obj->points[i].getVector3fMap();
                distance_obj = n.dot(y0) + p;

                // new cloud
                pcl::PointXYZI point_tmp;
                point_tmp.x = scene->points[indices[0]].x;
                point_tmp.y = scene->points[indices[0]].y;
                point_tmp.z = scene->points[indices[0]].z;

                if (distance_obj - distance > 0)
                    point_tmp.intensity = distance_obj - distance;
                else
                    point_tmp.intensity = 0;

                cloud_out->points.push_back(point_tmp);
            }

        }
        indices.clear();
        distances.clear();
    }

    cloud_out->width = cloud_out->size();
    cloud_out->height = 1;
    std::cout << "cloud_out size: " << cloud_out->size() << std::endl;
    return;
}

void WMSegmentation::computeDiffProjected(PointCloudT::Ptr obj, PointCloudT::Ptr scene,
                                          PointCloudIntT::Ptr cloud_output, PointCloudT::Ptr hull, pcl::ModelCoefficients::Ptr &plane)
{
    PointCloudT::Ptr scene_proj(new PointCloudT);
    PointCloudT::Ptr obj_proj(new PointCloudT);

    pcl::SACSegmentation<PointNT> seg;
    pcl::ModelCoefficients::Ptr coefficients_plane(new pcl::ModelCoefficients);
    pcl::PointIndices::Ptr inliers_plane(new pcl::PointIndices);

    seg.setOptimizeCoefficients(true);
    seg.setModelType(pcl::SACMODEL_PLANE);
    seg.setMethodType(pcl::SAC_RANSAC);
    seg.setMaxIterations(200);
    seg.setDistanceThreshold(0.05);
    seg.setInputCloud(obj);
    // Obtain the plane inliers and coefficients
    seg.segment(*inliers_plane, *coefficients_plane);
    //std::cerr << "Plane coefficients: " << *coefficients_plane << std::endl;

    PointCloudT::Ptr obj_filtered(new PointCloudT);
    removePerperndicularPlane(obj, coefficients_plane, obj_filtered);

    // Create the filtering object
    pcl::ProjectInliers<PointNT> proj;
    proj.setModelType(pcl::SACMODEL_PLANE);
    proj.setInputCloud(scene);
    proj.setModelCoefficients(coefficients_plane);
    proj.filter(*scene_proj);

    proj.setInputCloud(obj_filtered);
    proj.filter(*obj_proj);


    // Create a Concave Hull representation of the projected inliers
    pcl::ConcaveHull<PointNT> chull;
    chull.setInputCloud(obj_proj);
    chull.setAlpha(0.1);
    chull.reconstruct(*hull);

    getPointPlaneDistanceCloud(obj_filtered, scene, scene_proj, cloud_output, coefficients_plane);

    plane = coefficients_plane;
}

void WMSegmentation::removePerperndicularPlane(PointCloudT::Ptr &cloud, pcl::ModelCoefficients::Ptr &plane,
                                               PointCloudT::Ptr &cloud_filtered)
{

    // Creating the KdTree object for the search method of the extraction
    pcl::search::KdTree<PointNT>::Ptr tree(new pcl::search::KdTree<PointNT>);
    tree->setInputCloud(cloud);

    std::vector<pcl::PointIndices> cluster_indices;
    pcl::EuclideanClusterExtraction<PointNT> ec;
    ec.setClusterTolerance(0.01); // 2cm
    ec.setMinClusterSize(100);
    ec.setMaxClusterSize(25000);
    ec.setSearchMethod(tree);
    ec.setInputCloud(cloud);
    ec.extract(cluster_indices);

    // the index 0 is assumed to be the biggest so with the washing machine front side
    pcl::PointIndices::Ptr inliers(new pcl::PointIndices);
    inliers->indices = cluster_indices[0].indices;

    pcl::ExtractIndices<PointNT> extract;
    extract.setInputCloud(cloud);
    extract.setIndices(inliers);
    extract.setNegative(false);
    extract.filter(*cloud_filtered);
}

void WMSegmentation::segmentSceneCloud(PointCloudIntT::Ptr cloud, pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_seg, float threshold)
{
    pcl::PointXYZRGB pointTmp;
    for (int i = 0; i < cloud->size(); i++)
    {
        if (cloud->points[i].intensity > threshold)
        {
            pointTmp.x = cloud->points[i].x;
            pointTmp.y = cloud->points[i].y;
            pointTmp.z = cloud->points[i].z;
            cloud_seg->points.push_back(pointTmp);
        }
    }
}

bool WMSegmentation::compute(pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_seg)
{
    computeDiffProjected(object_, scene_, scene_map_, cloud_hull_, plane_);

    segmentSceneCloud(scene_map_, cloud_seg, threshold_seg_);

    return true;
}

void WMSegmentation::visualize(bool flag_spin)
{
    pcl::visualization::PCLVisualizer vizScene("scene_map_");
    vizScene.setBackgroundColor(1.0, 1.0, 1.0);
    vizScene.addPointCloud<PointNT>(cloud_hull_, "cloud_hull_");

    vizScene.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_COLOR, 0.7f, 0.0f, 0.0f, "cloud_hull_");
    pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity_distribution(scene_map_, "intensity");
    vizScene.addPointCloud<pcl::PointXYZI>(scene_map_, intensity_distribution, "diff_scene");

    if (flag_spin)
        vizScene.spin();
}
