#pragma once

#include <pcl/point_types.h>
#include <pcl/io/pcd_io.h>
#include <pcl/kdtree/kdtree_flann.h>
#include <pcl/surface/mls.h>

#include <pcl/visualization/cloud_viewer.h>
#include <pcl/filters/voxel_grid.h>
#include <pcl/filters/extract_indices.h>
#include <pcl/common/pca.h>
#include <pcl/segmentation/extract_clusters.h>
#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/common/distances.h>

#include <pcl/common/centroid.h>
#include <pcl/common/geometry.h>

#include <pcl/segmentation/supervoxel_clustering.h>

//VTK include needed for drawing graph lines
#include <vtkPolyLine.h>

#include <pcl/features/organized_edge_detection.h>

struct Spherical
{
    float inclination;
    float azimuth;
    float entropy;
    float entropy_normalized;
};

class EntropyFilter
{

private:
    pcl::PointCloud<pcl::PointXYZRGB>::Ptr m_source, m_cloud_downsample;
    pcl::PointCloud<pcl::PointNormal> m_mls_points;
    pcl::PointCloud<pcl::PointXYZ>::Ptr m_mls_cloud, m_edges_cloud;
    pcl::PointCloud<pcl::PointXYZI>::Ptr m_cloud_depth, m_cloud_convexity, m_cloud_seg, m_candidate_map;
    pcl::PointCloud<pcl::Normal>::Ptr m_mls_normals;
    pcl::PointCloud<Spherical>::Ptr m_spherical;

    pcl::ModelCoefficients::Ptr m_plane;

    float m_leafsize, m_entropy_threshold, m_curvature_threshold,
        m_depth_interval, m_depth_threshold, m_angle_threshold, _max_entropy, m_local_radius_entropy;
    int m_KNN;
    bool _flag_candidate_weight, _flag_depth_weight;
    int task_id_;

public:
    EntropyFilter() : m_source(new pcl::PointCloud<pcl::PointXYZRGB>),
                      m_cloud_downsample(new pcl::PointCloud<pcl::PointXYZRGB>),
                      m_mls_cloud(new pcl::PointCloud<pcl::PointXYZ>),
                      m_mls_normals(new pcl::PointCloud<pcl::Normal>),
                      m_spherical(new pcl::PointCloud<Spherical>),
                      m_cloud_seg(new pcl::PointCloud<pcl::PointXYZI>),
                      m_cloud_depth(new pcl::PointCloud<pcl::PointXYZI>),
                      m_cloud_convexity(new pcl::PointCloud<pcl::PointXYZI>),
                      m_candidate_map(new pcl::PointCloud<pcl::PointXYZI>),
                      m_plane(new pcl::ModelCoefficients),
                      m_edges_cloud(new pcl::PointCloud<pcl::PointXYZ>)

    {
        _flag_candidate_weight = false;
        _flag_depth_weight = false;
        m_KNN = 500;
        m_leafsize = 0.0;

        task_id_ = 1;
    }

    // set type of task considered: 1 (bin), 2 (washing machine), 3 (drum)
    void setTask(int task_id) {task_id_ = task_id; };

    pcl::PointCloud<pcl::PointXYZI>::Ptr getCloudSeg() { return m_cloud_seg; };

    pcl::PointCloud<pcl::PointXYZ>::Ptr getMlsCloud() { return m_mls_cloud; };

    pcl::PointCloud<pcl::Normal>::Ptr getMlsNormals() { return m_mls_normals; };

    void useWeightDepthMap(bool flag) { _flag_depth_weight = flag; }

    void useWeightCandidateMap(bool flag) { _flag_candidate_weight = flag; }

    void setReferencePlane(pcl::ModelCoefficients::Ptr &plane) { m_plane = plane; }

    void setInputCloud(pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud_in) { m_source = cloud_in; };

    void setDownsampleLeafSize(float leaf_size) { m_leafsize = leaf_size; };

    void setEntropyThreshold(float entropy_th) { m_entropy_threshold = entropy_th; };

    void setRadiusLocalSearchEntropy(float R) { m_local_radius_entropy = R; };

    void setCurvatureThreshold(float curvature_th) { m_curvature_threshold = curvature_th; };

    void setDepthThreshold(float depth_th) { m_depth_threshold = depth_th; };

    void setAngleThresholdForConvexity(float angle_th) { m_angle_threshold = angle_th; };

    void setEdgesCloud(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud) { m_edges_cloud = cloud; };

    pcl::ModelCoefficients::Ptr getReferencePlane() { return m_plane; };

    bool compute();

    // ---------------------------------------------------------------------------------
    //ColorMap functions
    void colorMapEntropy(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_map);

    void colorMapCurvature(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_map);

    void colorMapInclination(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_map);

    void colorMapAzimuth(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_map);

    //Visualization function
    void visualizeAll(bool flag);

private:
    void downsample(pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud_in,
                    float leaf_size, pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud_out);

    void computePolyFitting(pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud, pcl::PointCloud<pcl::PointNormal> &mls_points);

    void divideCloudNormals(pcl::PointCloud<pcl::PointNormal> &input, pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud,
                            pcl::PointCloud<pcl::Normal>::Ptr &normals);

    void getSpherical(pcl::PointCloud<pcl::Normal>::Ptr &cloud_normals, pcl::PointCloud<Spherical>::Ptr &spherical);

    void normalizeEntropy(pcl::PointCloud<Spherical>::Ptr &spherical);

    void histogram2D(pcl::PointCloud<Spherical>::Ptr &spherical, int id0, std::vector<int> indices, std::vector<float> distances);

    // LOCAL SEARCH
    void local_search(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, pcl::PointCloud<Spherical>::Ptr &spherical,
                      pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_combined,
                      pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_depth);

    void segmentCloudEntropy(pcl::PointCloud<pcl::PointNormal> &cloud, pcl::PointCloud<Spherical>::Ptr &spherical,
                             pcl::PointCloud<pcl::PointXYZI>::Ptr &output, float thresholdEntropy);

    float getPointPlaneDistanceCloud(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud,
                                     pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_out,
                                     pcl::ModelCoefficients::Ptr &coeff);

    void computeDepthMapInv(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_out);

    bool checkConvexity(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud,
                        pcl::PointCloud<pcl::Normal>::Ptr &normals,
                        int id0,
                        std::vector<int> ids);

    void splitPointNormal(pcl::PointCloud<pcl::PointNormal>::Ptr &input,
                          pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud,
                          pcl::PointCloud<pcl::Normal>::Ptr &normals);

    void localSearchForConvexity(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, pcl::PointCloud<pcl::Normal>::Ptr &normals,
                                 pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_map);

    void buildCandidateMap(pcl::PointCloud<pcl::PointXYZI>::Ptr &convexity,
                           pcl::PointCloud<pcl::Normal>::Ptr &normals,
                           pcl::PointCloud<pcl::PointXYZ>::Ptr &edges,
                           pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_map);
};
