#pragma once

#include <algorithm>

#include <pcl/ModelCoefficients.h>
#include <pcl/point_types.h>
#include <pcl/point_cloud.h>
#include <pcl/filters/project_inliers.h>
#include <pcl/common/centroid.h>
#include <pcl/common/common.h>
#include <pcl/common/eigen.h>
#include <pcl/common/pca.h>
#include <pcl/kdtree/kdtree_flann.h>
#include <pcl/segmentation/sac_segmentation.h>

#include <pcl/visualization/cloud_viewer.h>

typedef std::vector<Eigen::Affine3d, Eigen::aligned_allocator<Eigen::Affine3d>> Affine3dVector;
typedef std::vector<Eigen::Vector3f, Eigen::aligned_allocator<Eigen::Vector3f>> Vector3fVector;

struct FrameT
{
    pcl::PointXYZ x;
    pcl::PointXYZ y;
    pcl::PointXYZ z;
    pcl::PointXYZ o;
};

class Pose
{

private:
    pcl::PointCloud<pcl::PointXYZRGB>::Ptr source_;
    std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> splines_;
    pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_seg_;
    std::vector<pcl::PointCloud<pcl::PointXYZI>::Ptr> voxels_;

    pcl::ModelCoefficients::Ptr plane_;
    std::vector<FrameT> frames_;

    int task_id_;

public:
    Pose() : source_(new pcl::PointCloud<pcl::PointXYZRGB>),
             plane_(new pcl::ModelCoefficients),
             cloud_seg_(new pcl::PointCloud<pcl::PointXYZI>)

    {
        task_id_ = 1;
    }

    // set type of task considered: 1 (bin), 2 (washing machine), 3 (drum)
    void setTask(int task_id) {task_id_ = task_id; };

    void setSourceCloud(pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud) { source_ = cloud; };

    void setCloudSeg(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud) { cloud_seg_ = cloud; };

    void setPlane(pcl::ModelCoefficients::Ptr &plane) { plane_ = plane; };

    void setInputSplines(std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> &splines) { splines_ = splines; };

    void setInputVoxels(std::vector<pcl::PointCloud<pcl::PointXYZI>::Ptr> &cloud) { voxels_ = cloud; };

    void compute(Affine3dVector &transformation_matrix_vector);

    void visualize(bool flag_spin = false);

private:
    void project(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud,
                 pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_proj, pcl::ModelCoefficients::Ptr &plane);

    void project(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud,
                 pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_proj, pcl::ModelCoefficients::Ptr &plane);

    pcl::PointCloud<pcl::PointXYZI>::Ptr targetPoints(pcl::PointCloud<pcl::PointXYZI>::Ptr &voxel_cloud_proj,
                                                      pcl::PointCloud<pcl::PointXYZ>::Ptr &spline_cloud_proj);

    std::vector<Eigen::Vector3f> computeDirections(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_points,
                                                   pcl::PointCloud<pcl::PointXYZ>::Ptr &spline_cloud_proj);

    std::vector<Eigen::Vector3f> computeDirectionsDrum( pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_points,
                                                        std::vector<Eigen::Vector3f> &center_points,
                                                        pcl::PointCloud<pcl::PointXYZ>::Ptr &spline_cloud_proj);
    
    Eigen::Vector3f computeNormalDir(Eigen::Vector3f &frame_origin);

    std::vector<Eigen::Vector3f> computeCenters(pcl::PointCloud<pcl::PointXYZI>::Ptr &target_points,
                                                pcl::PointCloud<pcl::PointXYZ>::Ptr &spline_cloud_proj,
                                                pcl::PointCloud<pcl::PointXYZ>::Ptr &spline_cloud);

    Eigen::Affine3d computeTransformation(Eigen::Vector3f &origin, Eigen::Vector3f &directionX,
                                          Eigen::Vector3f &directionY, Eigen::Vector3f &directionZ);

    FrameT createFrameForVisualization(Eigen::Vector3f &w_center, Eigen::Vector3f &w_dir,
                                       Eigen::Vector3f &w_dir2, Eigen::Vector3f &w_dir3);

    void orderOutputMatirces(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_entropy, int task_id,
                                Affine3dVector &transformation_matrix_vector);
};