#include <pcl/console/parse.h>
#include <pcl/point_cloud.h>
#include <pcl/point_types.h>
#include <pcl/io/pcd_io.h>
#include <pcl/visualization/pcl_visualizer.h>
#include <pcl/segmentation/supervoxel_clustering.h>
#include <pcl/kdtree/kdtree_flann.h>
#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/filters/project_inliers.h>

//VTK include needed for drawing graph lines
#include <vtkPolyLine.h>

//#include "../include/von_mises_dist.h"

// Types
typedef pcl::PointXYZRGBA PointT2;
typedef pcl::PointCloud<PointT2> PointCloudT2;
typedef pcl::PointNormal PointNT;
typedef pcl::PointCloud<PointNT> PointNCloudT;
typedef pcl::PointXYZL PointLT;
typedef pcl::PointCloud<PointLT> PointLCloudT;

typedef struct
{
    double r, g, b;
} ColourT;

struct GraphT
{
    //pcl::PointXYZRGBA centroid;
    //std::vector<pcl::PointXYZRGBA> neighbors;
    size_t id;
    pcl::PointXYZI node;
    bool assigned;
    std::vector<size_t> neighor_ids;
    std::vector<pcl::PointXYZI> neighbor_nodes;
};

// Von Mises Distribution for curvature
class VonMises
{
private:
    double bessel_, kappa_, mu_;

public:

    VonMises()
    {
        kappa_ = 1;
        mu_ = M_PI;
        bessel_ = 1.26607; //boost::math::cyl_bessel_i(0, kappa_);
    }

    double pdf(double x)
    {
        return exp(kappa_ * cos(x - mu_)) / (2 * M_PI * bessel_);
    }
};


class SuperVoxel
{
private:
    PointCloudT2::Ptr cloud_, cloud_proj_;
    pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_entropy_, cloud_nodes_;
    pcl::PointCloud<pcl::Normal>::Ptr cloud_normals_;
    pcl::ModelCoefficients::Ptr plane_;
    VonMises vmises_;

    bool disable_transform_;
    float voxel_resolution_, seed_resolution_, color_importance_, spatial_importance_, normal_importance_;
    int task_id_;

    std::vector<GraphT> nodes_;
    std::vector<pcl::PointXYZI> curve_;
    std::vector<size_t> sequence_id_viz_;

public:
    SuperVoxel() : cloud_(new PointCloudT2),
                   cloud_entropy_(new pcl::PointCloud<pcl::PointXYZI>),
                   cloud_nodes_(new pcl::PointCloud<pcl::PointXYZI>),
                   cloud_normals_(new pcl::PointCloud<pcl::Normal>),
                   plane_(new pcl::ModelCoefficients)
    {
        disable_transform_ = true;
        voxel_resolution_ = 0.01f;
        seed_resolution_ = 0.1f;
        color_importance_ = 0.2f;
        spatial_importance_ = 0.8f;
        normal_importance_ = 0.0f;
        task_id_ = 1;
    }

    // set type of task considered: 1 (bin), 2 (washing machine), 3 (drum)
    void setTask(int task_id) {task_id_ = task_id; };

    void setEntropyCloud(pcl::PointCloud<pcl::PointXYZI>::Ptr cloud) { cloud_entropy_ = cloud; };

    void setNormalCloud(pcl::PointCloud<pcl::Normal>::Ptr cloud) { cloud_normals_ = cloud; };

    void setVoxelResolution(float val) { voxel_resolution_ = val; };

    void SetSeedResolution(float val) { seed_resolution_ = val; };

    void setColorImportance(float val) { color_importance_ = val; };

    void setSpatialImportance(float val) { spatial_importance_ = val; };

    void setNormalImportance(float val) { normal_importance_ = val; };

    void setInputPlane(pcl::ModelCoefficients::Ptr plane) { plane_ = plane; };

    // ------------
    size_t nodeMinNN(std::vector<GraphT> &nodes);
    size_t nnMaxIntensity(std::vector<size_t> &nn, std::vector<GraphT> &nodes);

    pcl::PointCloud<pcl::PointXYZRGBA>::Ptr projectCloud(pcl::PointCloud<pcl::PointXYZRGBA>::Ptr &cloud_in,
                                                         pcl::ModelCoefficients::Ptr &plane);
    pcl::PointCloud<pcl::PointXYZI>::Ptr translatePathToCloud(pcl::PointCloud<pcl::PointXYZRGBA>::Ptr &cloud,
                                                              pcl::PointCloud<pcl::PointXYZRGBA>::Ptr &cloud_proj,
                                                              pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_entropy,
                                                              std::vector<GraphT> &nodes,
                                                              std::vector<size_t> &path);

    std::vector<size_t> buildPath(std::vector<GraphT> &set, std::vector<GraphT> &nodes, std::vector<size_t> &path);

    void normalizeIntensity(std::vector<GraphT> &nodes);

    void combineScores(std::vector<GraphT> &nodes);

    bool computeSupervoxels(std::vector<pcl::PointCloud<pcl::PointXYZI>::Ptr> &clouds_output);

    std::vector<GraphT> populateGraph(std::map<std::uint32_t, pcl::Supervoxel<PointT2>::Ptr> supervoxel_clusters,
                                      std::multimap<std::uint32_t, std::uint32_t> supervoxel_adjacency);

    pcl::PointXYZI nnSearch(pcl::Supervoxel<PointT2>::Ptr supervoxel);

    pcl::PointCloud<pcl::PointXYZI>::Ptr getNodesCloud(std::vector<GraphT> nodes);

    void updateNodesIDs(std::vector<GraphT> &nodes);

    float computeCurvatureScore(std::vector<GraphT> &nodes, size_t last_id, size_t parent_id, size_t candidate_id);

    float vonmisesPdf(float angle);

    size_t nodeMaxIntensitySet(std::vector<GraphT> &set, std::vector<GraphT> &nodes);

    Eigen::Vector3f computeDirection(pcl::PointCloud<pcl::PointXYZRGBA>::Ptr cloud_voxel, Eigen::Vector3f centroid);

    pcl::PointCloud<pcl::PointXYZRGBA>::Ptr cloudVoxel(size_t target_id,
                                                       std::map<std::uint32_t, pcl::Supervoxel<PointT2>::Ptr> supervoxel_clusters,
                                                       std::multimap<std::uint32_t, std::uint32_t> supervoxel_adjacency);

    std::vector<float> neighborsScores(std::vector<GraphT> &set, std::vector<GraphT> &nodes,
                                       pcl::PointCloud<pcl::PointXYZRGBA>::Ptr &cloud_voxel,
                                       size_t index);

    int extractTwoNN(std::vector<float> &pdfs, std::vector<GraphT> &nodes, size_t &idx1, size_t &idx2);

    void displayNodesInformation(std::vector<GraphT> nodes);

    void splitSetNodes(std::vector<GraphT> &nodes, std::vector<std::vector<GraphT>> &sets);

    // visualization ------------------------------------------
    void visualization();
    void drawAllLines(pcl::visualization::PCLVisualizer::Ptr viz);


    // color cloud
    ColourT getColour(double v, double vmin, double vmax);
    pcl::PointCloud<pcl::PointXYZRGBA>::Ptr colorCloud(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud);


};