#include "../include/entropy.h"


bool EntropyFilter::compute()
{
    pcl::console::print_highlight("Inside EntropyFilter class...\n");

    computePolyFitting(m_source, m_mls_points);
    divideCloudNormals(m_mls_points, m_mls_cloud, m_mls_normals);
    getSpherical(m_mls_normals, m_spherical);
    pcl::console::print_highlight("Spherical coordinates computed.\n");

    if (task_id_ == 1 || task_id_ == 2) // bin or washing machine
    {
        m_depth_interval = getPointPlaneDistanceCloud(m_mls_cloud, m_cloud_depth, m_plane);
        if (fabs(m_depth_interval) < m_depth_threshold) 
        {
            PCL_WARN("Depth interval very small, grasp is not achievable! value: [%f mm]\n", m_depth_interval * 1000);
            return false;
        }
    }
    else if (task_id_ == 3) // drum
        computeDepthMapInv(m_mls_cloud, m_cloud_depth);  
    else
        PCL_WARN("Uknown task_id assigned!");

    pcl::console::print_highlight("Depth map computed.\n");

    localSearchForConvexity(m_mls_cloud, m_mls_normals, m_cloud_convexity);
    pcl::console::print_highlight("Convexity map computed.\n");

    buildCandidateMap(m_cloud_convexity, m_mls_normals, m_edges_cloud, m_candidate_map);
    pcl::console::print_highlight("Candidate map computed.\n");

    pcl::console::print_highlight("Building entropy map...\n");
    local_search(m_mls_cloud, m_spherical, m_candidate_map, m_cloud_depth);
    normalizeEntropy(m_spherical);
    pcl::console::print_highlight("Entropy map computed.\n");

    if (_max_entropy < 1.0)
    {
        PCL_WARN("Entropy very small!\n");
        //return false;
    }
    segmentCloudEntropy(m_mls_points, m_spherical, m_cloud_seg, m_entropy_threshold);
    pcl::console::print_highlight("Entropy map segmented, returning...\n");

    return true;
}

// -----------------------------------------------------------------------------------------------------------------------
//ColorMap functions
void EntropyFilter::colorMapEntropy(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_map)
{
    cloud_map->width = m_mls_cloud->width;
    cloud_map->height = m_mls_cloud->height;
    cloud_map->resize(m_mls_cloud->width * m_mls_cloud->height);

    for (int i = 0; i < m_mls_cloud->size(); i++)
    {
        cloud_map->points[i].x = m_mls_cloud->points[i].x;
        cloud_map->points[i].y = m_mls_cloud->points[i].y;
        cloud_map->points[i].z = m_mls_cloud->points[i].z;
        cloud_map->points[i].intensity = m_spherical->points[i].entropy;
    }
}

void EntropyFilter::colorMapCurvature(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_map)
{
    cloud_map->width = m_mls_cloud->width;
    cloud_map->height = m_mls_cloud->height;
    cloud_map->resize(m_mls_cloud->width * m_mls_cloud->height);

    for (int i = 0; i < m_mls_cloud->size(); i++)
    {
        cloud_map->points[i].x = m_mls_cloud->points[i].x;
        cloud_map->points[i].y = m_mls_cloud->points[i].y;
        cloud_map->points[i].z = m_mls_cloud->points[i].z;
        cloud_map->points[i].intensity = m_mls_normals->points[i].curvature;
    }
}

void EntropyFilter::colorMapInclination(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_map)
{
    cloud_map->width = m_mls_cloud->width;
    cloud_map->height = m_mls_cloud->height;
    cloud_map->resize(m_mls_cloud->width * m_mls_cloud->height);

    for (int i = 0; i < m_mls_cloud->size(); i++)
    {
        cloud_map->points[i].x = m_mls_cloud->points[i].x;
        cloud_map->points[i].y = m_mls_cloud->points[i].y;
        cloud_map->points[i].z = m_mls_cloud->points[i].z;
        cloud_map->points[i].intensity = m_spherical->points[i].inclination;
    }
}

void EntropyFilter::colorMapAzimuth(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_map)
{
    cloud_map->width = m_mls_cloud->width;
    cloud_map->height = m_mls_cloud->height;
    cloud_map->resize(m_mls_cloud->width * m_mls_cloud->height);

    for (int i = 0; i < m_mls_cloud->size(); i++)
    {
        cloud_map->points[i].x = m_mls_cloud->points[i].x;
        cloud_map->points[i].y = m_mls_cloud->points[i].y;
        cloud_map->points[i].z = m_mls_cloud->points[i].z;
        cloud_map->points[i].intensity = m_spherical->points[i].azimuth;
    }
}

void EntropyFilter::visualizeAll(bool flag)
{

    pcl::visualization::PCLVisualizer vizEdges("PCL Edges Cloud");
    vizEdges.setBackgroundColor(1.0f, 1.0f, 1.0f);
    vizEdges.addPointCloud<pcl::PointXYZRGB>(m_source, "m_source");
    vizEdges.addPointCloud<pcl::PointXYZ>(m_edges_cloud, "m_edges_cloud");
    vizEdges.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_COLOR, 0.0f, 1.0f, 0.0f, "m_edges_cloud");

    pcl::visualization::PCLVisualizer vizCandidate("Candidate Map");
    vizCandidate.setBackgroundColor(1.0f, 1.0f, 1.0f);
    pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity_distributionvCandidate(m_candidate_map, "intensity");
    vizCandidate.addPointCloud<pcl::PointXYZI>(m_candidate_map, intensity_distributionvCandidate, "m_candidate_map");
    vizCandidate.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 3, "m_candidate_map");

    pcl::visualization::PCLVisualizer vizSource("PCL Source Cloud");
    vizSource.setBackgroundColor(1.0f, 1.0f, 1.0f);
    vizSource.addPointCloud<pcl::PointXYZRGB>(m_source, "m_source");
    vizSource.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_COLOR, 0.0f, 1.0f, 0.0f, "m_source");

    pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_curvature(new pcl::PointCloud<pcl::PointXYZI>);
    colorMapCurvature(cloud_curvature);
    pcl::visualization::PCLVisualizer vizC("PCL Curvature Map");
    vizC.setBackgroundColor(1.0f, 1.0f, 1.0f);
    pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity_distributionCurvature(cloud_curvature, "intensity");
    vizC.addPointCloud<pcl::PointXYZI>(cloud_curvature, intensity_distributionCurvature, "cloud_mapCurvature");
    vizC.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 3, "cloud_mapCurvature");

    pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_entropy(new pcl::PointCloud<pcl::PointXYZI>);
    colorMapEntropy(cloud_entropy);
    pcl::visualization::PCLVisualizer vizE("PCL Entropy Map");
    vizE.setBackgroundColor(1.0f, 1.0f, 1.0f);
    pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity_distributionEntropy(cloud_entropy, "intensity");
    vizE.addPointCloud<pcl::PointXYZI>(cloud_entropy, intensity_distributionEntropy, "cloud_mapEntropy");
    vizE.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 3, "cloud_mapEntropy");

    if (flag)
    {
        pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_inclination(new pcl::PointCloud<pcl::PointXYZI>);
        colorMapInclination(cloud_inclination);
        pcl::visualization::PCLVisualizer vizI("PCL Inclination Map");
        vizI.setBackgroundColor(1.0f, 1.0f, 1.0f);
        pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity_distributionInclination(cloud_inclination, "intensity");
        vizI.addPointCloud<pcl::PointXYZI>(cloud_inclination, intensity_distributionInclination, "sample cloud_mapInclination");

        pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_azimuth(new pcl::PointCloud<pcl::PointXYZI>);
        colorMapAzimuth(cloud_azimuth);
        pcl::visualization::PCLVisualizer vizA("PCL Azimuth Map");
        vizA.setBackgroundColor(1.0f, 1.0f, 1.0f);
        pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity_distributionAzimuth(cloud_azimuth, "intensity");
        vizA.addPointCloud<pcl::PointXYZI>(cloud_azimuth, intensity_distributionAzimuth, "sample cloud_mapAzimuth");
    }

    pcl::visualization::PCLVisualizer vizDepth("PCL Depth Map");
    vizDepth.setBackgroundColor(1.0f, 1.0f, 1.0f);
    pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity_distributionDepth(m_cloud_depth, "intensity");
    vizDepth.addPointCloud<pcl::PointXYZI>(m_cloud_depth, intensity_distributionDepth, "cloud_depth");
    vizDepth.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 3, "cloud_depth");
    vizDepth.addPlane(*m_plane, "plane");

    pcl::visualization::PCLVisualizer vizConvexity("PCL Convexity Map");
    vizConvexity.setBackgroundColor(1.0f, 1.0f, 1.0f);
    pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity_distribution(m_cloud_convexity, "intensity");
    vizConvexity.addPointCloud<pcl::PointXYZI>(m_cloud_convexity, intensity_distribution, "cloud_convexity");
    vizConvexity.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 5, "cloud_convexity");
    //viz.addPointCloudNormals<pcl::PointXYZ, pcl::Normal>(cloud_small, normals_small, 1, 0.01);

    while (!vizC.wasStopped())
    {
        vizC.spinOnce(200);
    }
}
// -----------------------------------------------------------------------------------------------------------------------

void EntropyFilter::downsample(pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud_in, float leaf_size,
                               pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud_out)
{
    pcl::VoxelGrid<pcl::PointXYZRGB> vox_grid;
    vox_grid.setLeafSize(leaf_size, leaf_size, leaf_size);
    vox_grid.setInputCloud(cloud_in);
    vox_grid.filter(*cloud_out);
}

void EntropyFilter::computePolyFitting(pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud, pcl::PointCloud<pcl::PointNormal> &mls_points)
{
    // Output has the PointNormal type in order to store the normals calculated by MLS

    // Create a KD-Tree
    pcl::search::KdTree<pcl::PointXYZRGB>::Ptr tree(new pcl::search::KdTree<pcl::PointXYZRGB>);

    // Init object (second point type is for the normals, even if unused)
    pcl::MovingLeastSquares<pcl::PointXYZRGB, pcl::PointNormal> mls;

    mls.setComputeNormals(true);

    // Set parameters
    mls.setNumberOfThreads(4);
    mls.setInputCloud(cloud);
    mls.setPolynomialOrder(2);
    mls.setSearchMethod(tree);
    mls.setSearchRadius(0.03);

    // Reconstruct
    mls.process(mls_points);
    std::cout << "mls_points: " << mls_points.height << ", " << mls_points.width << std::endl;
}

void EntropyFilter::divideCloudNormals(pcl::PointCloud<pcl::PointNormal> &input, pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud,
                                       pcl::PointCloud<pcl::Normal>::Ptr &normals)
{
    cloud->height = input.height;
    cloud->width = input.width;
    cloud->resize(input.size());
    normals->height = input.height;
    normals->width = input.width;
    normals->resize(input.size());
    for (int i = 0; i < input.size(); i++)
    {
        cloud->points[i].x = input.points[i].x;
        cloud->points[i].y = input.points[i].y;
        cloud->points[i].z = input.points[i].z;

        normals->points[i].normal_x = input.points[i].normal_x;
        normals->points[i].normal_y = input.points[i].normal_y;
        normals->points[i].normal_z = input.points[i].normal_z;
        normals->points[i].curvature = input.points[i].curvature;
    }
}

void EntropyFilter::getSpherical(pcl::PointCloud<pcl::Normal>::Ptr &cloud_normals, pcl::PointCloud<Spherical>::Ptr &spherical)
{
    spherical->width = cloud_normals->width;
    spherical->height = cloud_normals->height;
    spherical->resize(spherical->width * spherical->height);

    pcl::Normal data;
    for (size_t i = 0; i < cloud_normals->points.size(); ++i)
    {
        data = cloud_normals->points[i];
        spherical->points[i].azimuth = atan2(data.normal_z, data.normal_y);
        spherical->points[i].inclination = atan2(sqrt(data.normal_y * data.normal_y + data.normal_z * data.normal_z), data.normal_x);
    }
}

void EntropyFilter::normalizeEntropy(pcl::PointCloud<Spherical>::Ptr &spherical)
{

    float max_entropy = 0;
    for (int i = 0; i < spherical->size(); i++)
    {
        if (spherical->points[i].entropy > max_entropy)
            max_entropy = spherical->points[i].entropy;
    }

    for (int i = 0; i < spherical->size(); i++)
    {
        spherical->points[i].entropy_normalized = spherical->points[i].entropy / max_entropy;
    }
    std::cout << "max entropy : " << max_entropy << std::endl;
    _max_entropy = max_entropy;
}

//LOCAL HISTOGRAM and entropy calculation at the end.
void EntropyFilter::histogram2D(pcl::PointCloud<Spherical>::Ptr &spherical, int id0, std::vector<int> indices, std::vector<float> distances)
{

    int Hist[64][64] = {0};
    float step_inc = M_PI / 63;
    float step_az = (2 * M_PI) / 63;
    long bin_inc, bin_az;
    for (int i = 0; i < indices.size(); i++)
    {
        bin_inc = std::lroundf(spherical->points[indices[i]].inclination / step_inc);
        bin_az = std::lroundf((spherical->points[indices[i]].azimuth + M_PI) / step_az);
        //if (bin_az < 0)
        //    std::cout << "erorr bin_az negative" << std::endl;
        Hist[bin_inc][bin_az]++;
    }
    bin_inc = std::lroundf(spherical->points[id0].inclination / step_inc);
    bin_az = std::lroundf((spherical->points[id0].azimuth + M_PI) / step_az);
    //if (bin_az < 0)
    //    std::cout << "erorr bin_az negative" << std::endl;
    Hist[bin_inc][bin_az]++;

    // ----------------------------------------------
    // histogram normalization
    float HistNorm[64][64] = {0};
    float max_value = 0;
    for (int i = 0; i < 64; i++)
        for (int j = 0; j < 64; j++)
            if (Hist[i][j] > max_value)
                max_value = Hist[i][j];

    for (int i = 0; i < 64; i++)
        for (int j = 0; j < 64; j++)
            HistNorm[i][j] = Hist[i][j] / max_value;

    // ----------------------------------------------
    //entropy calculation
    float entropy_value = 0;
    float temp_entropy = 0;
    for (int i = 0; i < 64; i++)
        for (int j = 0; j < 64; j++)
        {
            temp_entropy = HistNorm[i][j] * log2(HistNorm[i][j]);

            if (!std::isnan(temp_entropy))
                entropy_value += temp_entropy;
        }

    spherical->points[id0].entropy = fabs(entropy_value);
    //std::cout << "entropy value: " << spherical->points[id0].entropy << std::endl;
}

// LOCAL SEARCH
void EntropyFilter::local_search(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, pcl::PointCloud<Spherical>::Ptr &spherical,
                                 pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_combined,
                                 pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_depth)
{
    pcl::KdTreeFLANN<pcl::PointXYZ> kdtree;
    kdtree.setInputCloud(cloud);
    pcl::PointXYZ searchPoint;

    for (int it = 0; it < cloud->points.size(); it++)
    {

        if (cloud_combined->points[it].intensity >= 0.5)
        {

            searchPoint.x = cloud->points[it].x;
            searchPoint.y = cloud->points[it].y;
            searchPoint.z = cloud->points[it].z;

            std::vector<int> indices;
            std::vector<float> distances;

            if (kdtree.radiusSearch(searchPoint, m_local_radius_entropy, indices, distances) > 0)
            {
                histogram2D(spherical, it, indices, distances);
            }
            if (_flag_candidate_weight)
                spherical->points[it].entropy *= cloud_combined->points[it].intensity;
            if (_flag_depth_weight)
                spherical->points[it].entropy *= cloud_depth->points[it].intensity;
        }
        else
        {
            spherical->points[it].entropy = 0;
        }
    }
}

void EntropyFilter::segmentCloudEntropy(pcl::PointCloud<pcl::PointNormal> &cloud, pcl::PointCloud<Spherical>::Ptr &spherical,
                                        pcl::PointCloud<pcl::PointXYZI>::Ptr &output, float thresholdEntropy)
{
    pcl::PointXYZI p;
    for (int i = 0; i < spherical->size(); i++)
    {
        if (spherical->points[i].entropy_normalized > thresholdEntropy)
        {
            p.x = cloud.points[i].x;
            p.y = cloud.points[i].y;
            p.z = cloud.points[i].z;
            p.intensity = spherical->points[i].entropy_normalized;
            output->points.push_back(p);
        }
    }
    std::cout << "cloud segmented size: " << output->size() << std::endl;
}

float EntropyFilter::getPointPlaneDistanceCloud(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud,
                                                pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_out,
                                                pcl::ModelCoefficients::Ptr &plane)
{
    // output cloud initialization
    cloud_out->width = cloud->width;
    cloud_out->height = cloud->height;
    cloud_out->resize(cloud_out->width * cloud_out->height);

    // plane parameters
    Eigen::Vector3f n(plane->values[0], plane->values[1], plane->values[2]);
    float p = plane->values[3];

    // shift plane few centimeters below
    if (p > 0) // the origin is located on the side of vector "n"
        p += 0.08;
    else // the origin is located on the opposite side
        p -= 0.08;

    // distance computation
    Eigen::Vector3f x0;
    float distance;
    std::vector<float> v_distances;
    for (int i = 0; i < cloud->size(); i++)
    {
        // D=n·x_0+p

        if ((!std::isnan(cloud->points[i].x)) && (!std::isnan(cloud->points[i].y)) && (!std::isnan(cloud->points[i].z)))
        {
            cloud_out->points[i].x = cloud->points[i].x;
            cloud_out->points[i].y = cloud->points[i].y;
            cloud_out->points[i].z = cloud->points[i].z;

            x0 = cloud->points[i].getVector3fMap();
            distance = n.dot(x0) + p;           

            cloud_out->points[i].intensity = distance;
            //std::cout << distance << std::endl;
            v_distances.push_back(distance);
        }
    }

    /*
    pcl::visualization::PCLVisualizer viz("PCL cloud_out");
    viz.addCoordinateSystem(0.1, "coord", 0);
    viz.addPointCloud(cloud, "cloud");
    pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity_distributionvCandidate(cloud_out, "intensity");
    viz.addPointCloud<pcl::PointXYZI>(cloud_out, intensity_distributionvCandidate, "cloud_out");
    viz.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 3, "cloud_out");

    viz.addPlane(*plane, "plane");
    viz.spin();
    */

    float depth_max = *std::max_element(v_distances.begin(), v_distances.end());
    float depth_min = *std::min_element(v_distances.begin(), v_distances.end());

    std::cout << "depth min and max: " << depth_min << ", " << depth_max << std::endl;

    // all the points should have a positive or negative value; the plane should not cross them!
    if ((depth_max > 0 && depth_min < 0) || (depth_max < 0 && depth_min > 0))
    {
        PCL_WARN("Inconsistent signs\n");
        return 0;
    }
    std::cout << "depth ok" << std::endl;

    // normalization between 0 and 1; if negative value we need to invert the scale
    for (int i = 0; i < cloud_out->size(); i++)
        if (depth_max > 0)
            cloud_out->points[i].intensity = (cloud_out->points[i].intensity - depth_min) / fabs(depth_max - depth_min);
        else
            cloud_out->points[i].intensity = 1 - (cloud_out->points[i].intensity - depth_min) / fabs(depth_max - depth_min);

    return fabs(depth_max - depth_min);
}


// only for the drum use_case
void EntropyFilter::computeDepthMapInv(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_out)
{
    cloud_out->width = cloud->width;
    cloud_out->height = cloud->height;
    cloud_out->resize(cloud_out->width * cloud_out->height);

    double sqr_norm, value;

    Eigen::Vector4f line_pt = {0, 0, 0, 0};
    Eigen::Vector4f line_dir = {1, 0, 0, 0};

    sqr_norm = sqrt(line_dir.norm());

    for (int k = 0; k < cloud->size(); k++)
    {
        value = pcl::sqrPointToLineDistance(cloud->points[k].getVector4fMap(), line_pt, line_dir, sqr_norm);

        cloud_out->points[k].x = cloud->points[k].x;
        cloud_out->points[k].y = cloud->points[k].y;
        cloud_out->points[k].z = cloud->points[k].z;

        cloud_out->points[k].intensity = 1 - sqrt(value);
    }
}



bool EntropyFilter::checkConvexity(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, pcl::PointCloud<pcl::Normal>::Ptr &normals,
                                   int id0, std::vector<int> ids)
{
    Eigen::Vector3f x1 = cloud->points[id0].getVector3fMap();
    Eigen::Vector3f n1 = -1 * normals->points[id0].getNormalVector3fMap();

    Eigen::Vector3f d;

    float result;
    int convex = 0, concave = 0;
    for (int i = 0; i < ids.size(); i++)
    {
        Eigen::Vector3f x2 = cloud->points[ids[i]].getVector3fMap();
        Eigen::Vector3f n2 = -1 * normals->points[ids[i]].getNormalVector3fMap();
        d = x1 - x2;

        float angle = acos(n1.dot(n2)) * (180 / M_PI);
        if (angle > m_angle_threshold || angle < -m_angle_threshold) // if normals are similar we do not check convexity/concavity
        {
            //std::cout << "angle " << angle << std::endl;

            result = n1.dot(d) - n2.dot(d);
            if (result > 0)
                convex++;
            else
                concave++;
        }
    }

    //std::cout << "convex: " << convex << "; concave: " << concave << std::endl;

    if (convex > 0)
        return true;
    else
        return false;
}

void EntropyFilter::splitPointNormal(pcl::PointCloud<pcl::PointNormal>::Ptr &input, pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud,
                                     pcl::PointCloud<pcl::Normal>::Ptr &normals)
{
    cloud->height = input->height;
    cloud->width = input->width;
    cloud->resize(input->size());
    normals->height = input->height;
    normals->width = input->width;
    normals->resize(input->size());
    for (int i = 0; i < input->size(); i++)
    {
        cloud->points[i].x = input->points[i].x;
        cloud->points[i].y = input->points[i].y;
        cloud->points[i].z = input->points[i].z;

        normals->points[i].normal_x = input->points[i].normal_x;
        normals->points[i].normal_y = input->points[i].normal_y;
        normals->points[i].normal_z = input->points[i].normal_z;
        normals->points[i].curvature = input->points[i].curvature;
    }
}

void EntropyFilter::localSearchForConvexity(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, pcl::PointCloud<pcl::Normal>::Ptr &normals,
                                            pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_map)
{

    /*
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZ>);
    pcl::PointCloud<pcl::Normal>::Ptr normals(new pcl::PointCloud<pcl::Normal>);
    splitPointNormal(cloud_ready, cloud, normals);
    */
    pcl::KdTreeFLANN<pcl::PointXYZ> kdtree;
    kdtree.setInputCloud(cloud);
    pcl::PointXYZ searchPoint;
    int K = 20;

    cloud_map->width = cloud->width;
    cloud_map->height = cloud->height;
    cloud_map->resize(cloud->width * cloud->height);

    for (size_t it = 0; it < cloud->points.size(); it++)
    {
        searchPoint.x = cloud->points[it].x;
        searchPoint.y = cloud->points[it].y;
        searchPoint.z = cloud->points[it].z;

        // K nearest neighbor search
        std::vector<int> pointIdxNKNSearch(K);
        std::vector<float> pointNKNSquaredDistance(K);

        if (kdtree.nearestKSearch(searchPoint, K, pointIdxNKNSearch, pointNKNSquaredDistance) > 0)
        {
            if (checkConvexity(cloud, normals, it, pointIdxNKNSearch) == true)
                cloud_map->points[it].intensity = 1.0;
            else
                cloud_map->points[it].intensity = 0.0;

            cloud_map->points[it].x = cloud->points[it].x;
            cloud_map->points[it].y = cloud->points[it].y;
            cloud_map->points[it].z = cloud->points[it].z;
        }
    }
}

void EntropyFilter::buildCandidateMap(pcl::PointCloud<pcl::PointXYZI>::Ptr &convexity,
                                      pcl::PointCloud<pcl::Normal>::Ptr &normals,
                                      pcl::PointCloud<pcl::PointXYZ>::Ptr &edges,
                                      pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_map)
{
    cloud_map->width = convexity->width;
    cloud_map->height = convexity->height;
    cloud_map->resize(convexity->width * convexity->height);

    for (std::size_t i = 0; i < normals->points.size(); ++i)
    {
        cloud_map->points[i].x = convexity->points[i].x;
        cloud_map->points[i].y = convexity->points[i].y;
        cloud_map->points[i].z = convexity->points[i].z;
        cloud_map->points[i].intensity = 0;

        if (normals->points[i].curvature >= m_curvature_threshold)
            cloud_map->points[i].intensity += 1;

        if (convexity->points[i].intensity > 0)
            cloud_map->points[i].intensity += 1;
    }

    // if edges cloud has been provided
    if (edges->size() != 0)
    {
        pcl::KdTreeFLANN<pcl::PointXYZ> kdtree;
        kdtree.setInputCloud(edges);
        pcl::PointXYZ searchPoint;
        std::vector<int> indices;
        std::vector<float> distances;
        float radius = 0.01;

        for (std::size_t i = 0; i < cloud_map->points.size(); ++i)
        {
            searchPoint.getVector3fMap() = cloud_map->points[i].getVector3fMap();
            if (kdtree.radiusSearch(searchPoint, radius, indices, distances) > 0)
                cloud_map->points[i].intensity += 1;
        }
    }

    // normalize
    float max = 0;
    for (int i = 0; i < cloud_map->size(); i++)
    {
        if (cloud_map->points[i].intensity > max)
            max = cloud_map->points[i].intensity;
    }

    for (int i = 0; i < cloud_map->size(); i++)
        cloud_map->points[i].intensity /= max;
}