#include "ros/ros.h"
#include <geometry_msgs/Pose.h>
#include <sensor_msgs/PointCloud2.h>
#include <eigen_conversions/eigen_msg.h>
#include <tf2_ros/transform_listener.h>
#include <tf2_geometry_msgs/tf2_geometry_msgs.h>
#include <tf2_ros/static_transform_broadcaster.h>
#include <geometry_msgs/TransformStamped.h>
#include <tf2/LinearMath/Quaternion.h>
#include <std_msgs/Float64MultiArray.h>
#include "pcl_ros/point_cloud.h"
#include <pcl/common/transforms.h>
#include <pcl/filters/passthrough.h>
#include <Eigen/Geometry>
#include "tf2_eigen/tf2_eigen.h"

#include "../include/entropy.h"
#include "../include/supervoxel.h"
#include "../include/spline.h"
#include "../include/spline_pose.h"


#include "vision_msgs/WrinkledPosesEstimator.h"


class Vision{

    std::string topic_srv_;
    std::string target_frame_;
    std::string source_frame_bin_, source_frame_wm_;

    float entropy_th_, radius_search_, curvature_th_, depth_th_, convexity_th_;
    bool use_candidate_map_, use_depth_map_;

    ros::NodeHandle n_;
    ros::ServiceServer get_pose_srv_;
    ros::Subscriber sub_cloud_;
    tf2_ros::Buffer tfBuffer_;
    tf2_ros::TransformListener* tfListener_; 
    std::vector<geometry_msgs::Pose> poses_world_;

    pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_;

public:
    Vision(ros::NodeHandle* n);

    bool pipeline(  pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud, int task_id,
                    pcl::ModelCoefficients::Ptr &plane,
                    std::vector<geometry_msgs::Pose> &poses);

    bool serviceCallback(   vision_msgs::WrinkledPosesEstimator::Request &req,
                            vision_msgs::WrinkledPosesEstimator::Response &res);
};

Vision::Vision(ros::NodeHandle* n):n_(*n), cloud_(new pcl::PointCloud<pcl::PointXYZRGB>)
{
    std::cout<<"Constructor started"<<std::endl;

    n_.getParam("wrinkled_poses/topic_srv", topic_srv_);
    n_.getParam("wrinkled_poses/target_frame", target_frame_);
    n_.getParam("wrinkled_poses/source_frame_bin", source_frame_bin_);
    n_.getParam("wrinkled_poses/source_frame_wm", source_frame_wm_);

    n_.getParam("wrinkled_poses/entropy_th", entropy_th_);
    n_.getParam("wrinkled_poses/radius_search", radius_search_);
    n_.getParam("wrinkled_poses/curvature_th", curvature_th_);
    n_.getParam("wrinkled_poses/depth_th", depth_th_);
    n_.getParam("wrinkled_poses/use_depth_map", use_depth_map_);
    n_.getParam("wrinkled_poses/use_candidate_map", use_candidate_map_);

    get_pose_srv_ = n_.advertiseService(topic_srv_, &Vision::serviceCallback, this);
    
    tfListener_ = new tf2_ros::TransformListener(tfBuffer_);

    std::cout<<"Constructor finished"<<std::endl;
}


bool Vision::serviceCallback(vision_msgs::WrinkledPosesEstimator::Request &req, vision_msgs::WrinkledPosesEstimator::Response &res){


    std::cout << "cloud_seg size: " << req.cloud_seg.height * req.cloud_seg.width << std::endl;

    // PREPARE INPUT DATA

    // cloud
    pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZRGB>);
    pcl::fromROSMsg(req.cloud_seg, *cloud);

    // plane
    std::vector<double> req_plane = req.plane;
    pcl::ModelCoefficients::Ptr plane(new pcl::ModelCoefficients);
    plane->values.push_back(req_plane[0]);
    plane->values.push_back(req_plane[1]);
    plane->values.push_back(req_plane[2]);
    plane->values.push_back(req_plane[3]);

    // task_id
    int task_id = req.task_id;   
    
    // ------------------------------------
    // WORLD-CAMERA TRANSFORM
    std::string source_frame;
    if (task_id == 1)
        source_frame = source_frame_bin_;
    else
        source_frame = source_frame_wm_;

    geometry_msgs::TransformStamped transform;
    try{
        transform = tfBuffer_.lookupTransform(target_frame_, source_frame, ros::Time(0), ros::Duration(3));
    }
    catch (tf2::TransformException &ex){
        ROS_WARN("%s", ex.what());
    }

    // ------------------------------------
    // CALL TO VISION PIPELINE
    std::cout << "Input cloud size: "<< cloud->size() << std::endl;
    std::cout << "Task id: " << task_id << std::endl;
    std::vector<geometry_msgs::Pose> poses;
    if (!pipeline(cloud, task_id, plane, poses)) 
        ROS_WARN("ERROR WRINKLED POSES ESTIMATOR CALLBACK");
    

    // ------------------------------------
    // FROM CAMERA TO WORLD FRAME
    poses_world_.clear();
    for(int i = 0; i < poses.size(); i++){
        geometry_msgs::Pose pose_out;
        tf2::doTransform(poses[i], pose_out, transform);
        poses_world_.push_back(pose_out);

        ROS_WARN("Position: %f, %f, %f", poses[i].position.x, poses[i].position.y, poses[i].position.z);
        ROS_WARN("Orientation: %f, %f, %f, %f", poses[i].orientation.x, poses[i].orientation.y, poses[i].orientation.z, poses[i].orientation.w);

    }

    // ------------------------------------
    geometry_msgs::PoseArray temp;
    for(int i = 0; i<poses_world_.size(); i++)
    {
        temp.poses.push_back(poses_world_[i]);
    }
        
    res.poses = temp;
    return true;
}


bool Vision::pipeline(  pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud, int task_id,
                        pcl::ModelCoefficients::Ptr &plane,
                        std::vector<geometry_msgs::Pose> &poses)
{

   
    // ----------- ENTROPY FILTER -----------
    pcl::console::print_highlight("#### Entropy Filter...\n");
    EntropyFilter ef;
    ef.setTask(task_id); 
    ef.setInputCloud(cloud);
    ef.setReferencePlane(plane);
    //ef.setEdgesCloud(cloud_edges);
    //ef.setDownsampleLeafSize(EF_LEAF_SIZE);
    ef.setEntropyThreshold(entropy_th_);
    ef.setRadiusLocalSearchEntropy(radius_search_);
    ef.setCurvatureThreshold(curvature_th_);
    ef.setDepthThreshold(depth_th_);
    ef.setAngleThresholdForConvexity(convexity_th_);
    ef.useWeightDepthMap(use_depth_map_);
    ef.useWeightCandidateMap(use_candidate_map_);

    if (!ef.compute())
        return -1;

    pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_seg(new pcl::PointCloud<pcl::PointXYZI>);
    cloud_seg = ef.getCloudSeg();

    // ----------- WRINKLES DETECTOR -----------
    pcl::console::print_highlight("#### Supervoxels...\n");
    SuperVoxel svoxel;
    svoxel.setTask(task_id); 
    svoxel.setEntropyCloud(cloud_seg);
    svoxel.setInputPlane(plane);
    svoxel.setVoxelResolution(0.01);
    svoxel.SetSeedResolution(0.05);
    svoxel.setColorImportance(0.7f);
    svoxel.setSpatialImportance(0.9f);

    std::vector<pcl::PointCloud<pcl::PointXYZI>::Ptr> clouds_svoxel;
    svoxel.computeSupervoxels(clouds_svoxel);

    pcl::console::print_highlight("#### Spline...\n");
    Spline spline;
    std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> splines_result;
    for (auto cloud_sv : clouds_svoxel)
        splines_result.push_back(spline.compute(cloud_sv));

    pcl::console::print_highlight("#### Pose out...\n");
    Pose spose;
    spose.setTask(task_id); 
    spose.setSourceCloud(cloud);
    spose.setCloudSeg(cloud_seg);
    spose.setInputSplines(splines_result);
    spose.setInputVoxels(clouds_svoxel);
    spose.setPlane(plane);

    std::vector<Eigen::Affine3d, Eigen::aligned_allocator<Eigen::Affine3d>> matrices;
    spose.compute(matrices);

    for(int i = 0; i < matrices.size(); i++){

        std::cout << matrices[i].matrix() << std::endl;

        geometry_msgs::Pose output;
        tf::poseEigenToMsg(matrices[i], output);
        poses.push_back(output);

        ROS_INFO("Position: %f, %f, %f", output.position.x, output.position.y, output.position.z);
        ROS_INFO("Orientation: %f, %f, %f, %f",
             output.orientation.x, output.orientation.y, output.orientation.z, output.orientation.w);

    }
    pcl::console::print_highlight("#### DONE ####\n");
    return true;
}




int main(int argc, char **argv)
{
    ros::init(argc, argv, "wrinkled_poses_estimator");
    ros::NodeHandle n;
    Vision v(&n);
    ros::spin();
    return 0;
}
