#include "../include/spline_pose.h"

void Pose::project(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud,
                   pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_proj, pcl::ModelCoefficients::Ptr &plane)
{
    pcl::ProjectInliers<pcl::PointXYZ> proj;
    proj.setModelType(pcl::SACMODEL_PLANE);
    proj.setInputCloud(cloud);
    proj.setModelCoefficients(plane);
    proj.filter(*cloud_proj);
}

void Pose::project(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud,
                   pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_proj, pcl::ModelCoefficients::Ptr &plane)
{
    pcl::ProjectInliers<pcl::PointXYZI> proj;
    proj.setModelType(pcl::SACMODEL_PLANE);
    proj.setInputCloud(cloud);
    proj.setModelCoefficients(plane);
    proj.filter(*cloud_proj);
}

void Pose::compute(Affine3dVector &transformation_matrix_vector)
{

    if (splines_.size() != voxels_.size())
        std::cout << "splines and voxels vectors of different sizes!" << std::endl;


    Eigen::Vector3f normal_dir;
    if (task_id_ == 1 || task_id_ == 2) // bin or wm
    {
        normal_dir = {plane_->values[0], plane_->values[1], plane_->values[2]};
        normal_dir.normalize();
        if (plane_->values[3] > 0) // it is positive if the normal points toward the origin, we need the opposite! 
            normal_dir = -normal_dir;
    }

    std::vector<FrameT> frames;
    Eigen::Affine3d matrix;
    FrameT frame;
    for (size_t i = 0; i < splines_.size(); ++i)
    {
        Eigen::Vector3f w_dir1, w_dir2, w_dir3, center_point;
        pcl::PointCloud<pcl::PointXYZI>::Ptr target_points;
        std::vector<Eigen::Vector3f> cps, dirs;

        if (task_id_ == 1 || task_id_ == 2) // bin or wm
        {
            pcl::PointCloud<pcl::PointXYZ>::Ptr spline_cloud_proj(new pcl::PointCloud<pcl::PointXYZ>);
            project(splines_[i], spline_cloud_proj, plane_);

            pcl::PointCloud<pcl::PointXYZI>::Ptr voxel_cloud_proj(new pcl::PointCloud<pcl::PointXYZI>);
            project(voxels_[i], voxel_cloud_proj, plane_);

            target_points = targetPoints(voxel_cloud_proj, spline_cloud_proj);
            cps = computeCenters(target_points, spline_cloud_proj, splines_[i]);
            dirs = computeDirections(target_points, spline_cloud_proj);
        }
        else if (task_id_ == 3) // drum
        {
            target_points = targetPoints(voxels_[i], splines_[i]);
            cps = computeCenters(target_points, splines_[i], splines_[i]);
            dirs = computeDirectionsDrum(target_points, cps, splines_[i]);
        }
        else
            std::cout << "task_id not know!" << std::endl;


        for (size_t k = 0; k < dirs.size(); ++k)
        {
            center_point = cps[k];
            w_dir3 = dirs[k];
            w_dir3.normalize();

            if (task_id_ == 1 || task_id_ == 2) // bin or wm
                w_dir1 = normal_dir;
            else if (task_id_ == 3) // drum
                w_dir1 = computeNormalDir(center_point);
            else
                std::cout << "task_id not know!" << std::endl;
            w_dir1.normalize();

            w_dir2 = w_dir1.cross(w_dir3);

            frame = createFrameForVisualization(center_point, w_dir3, w_dir2, w_dir1);
            matrix = computeTransformation(center_point, w_dir3, w_dir2, w_dir1);

            frames_.push_back(frame);
            transformation_matrix_vector.push_back(matrix);
        }

        orderOutputMatirces(cloud_seg_, task_id_, transformation_matrix_vector);   
    }
}

pcl::PointCloud<pcl::PointXYZI>::Ptr Pose::targetPoints(pcl::PointCloud<pcl::PointXYZI>::Ptr &voxel_cloud_proj,
                                                        pcl::PointCloud<pcl::PointXYZ>::Ptr &spline_cloud_proj)
{

    pcl::KdTreeFLANN<pcl::PointXYZ> kdtree;
    kdtree.setInputCloud(spline_cloud_proj);
    int K = 1;
    std::vector<int> indices(K);
    std::vector<float> distances(K);
    pcl::PointXYZ s_point;
    pcl::PointXYZI t_point;

    pcl::PointCloud<pcl::PointXYZI>::Ptr target_points(new pcl::PointCloud<pcl::PointXYZI>);
    target_points->reserve(spline_cloud_proj->size());

    for (int i = 0; i < voxel_cloud_proj->size(); ++i)
    {
        s_point.x = voxel_cloud_proj->points[i].x;
        s_point.y = voxel_cloud_proj->points[i].y;
        s_point.z = voxel_cloud_proj->points[i].z;
        if (kdtree.nearestKSearch(s_point, K, indices, distances) > 0)
        {
            t_point.x = spline_cloud_proj->points[indices[0]].x;
            t_point.y = spline_cloud_proj->points[indices[0]].y;
            t_point.z = spline_cloud_proj->points[indices[0]].z;
            t_point.intensity = voxel_cloud_proj->points[i].intensity;

            target_points->points.push_back(t_point);
        }
        indices.clear();
        distances.clear();
    }

    return target_points;
}

std::vector<Eigen::Vector3f> Pose::computeCenters(pcl::PointCloud<pcl::PointXYZI>::Ptr &target_points,
                                                  pcl::PointCloud<pcl::PointXYZ>::Ptr &spline_cloud_proj,
                                                  pcl::PointCloud<pcl::PointXYZ>::Ptr &spline_cloud)
{
    std::vector<Eigen::Vector3f> centers;
    centers.reserve(target_points->size());

    pcl::KdTreeFLANN<pcl::PointXYZ> kdtree;
    kdtree.setInputCloud(spline_cloud_proj);
    int K = 1;
    std::vector<int> indices(K);
    std::vector<float> distances(K);
    pcl::PointXYZ s_point;

    Eigen::Vector3f c;
    for (size_t i = 0; i < target_points->size(); ++i)
    {
        s_point.x = target_points->points[i].x;
        s_point.y = target_points->points[i].y;
        s_point.z = target_points->points[i].z;

        if (kdtree.nearestKSearch(s_point, K, indices, distances) > 0)
        {
            c = spline_cloud->points[indices[0]].getVector3fMap();
            centers.push_back(c);
        }
    }
    return centers;
}

std::vector<Eigen::Vector3f> Pose::computeDirections(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_points,
                                                     pcl::PointCloud<pcl::PointXYZ>::Ptr &spline_cloud_proj)
{
    std::vector<Eigen::Vector3f> directions;
    directions.reserve(cloud_points->size());

    pcl::KdTreeFLANN<pcl::PointXYZ> kdtree;
    kdtree.setInputCloud(spline_cloud_proj);
    int K = 2;
    std::vector<int> indices(K);
    std::vector<float> distances(K);
    pcl::PointXYZ s_point;

    Eigen::Vector3f dir;
    for (size_t i = 0; i < cloud_points->size(); ++i)
    {
        s_point.x = cloud_points->points[i].x;
        s_point.y = cloud_points->points[i].y;
        s_point.z = cloud_points->points[i].z;

        // second point on the spline_cloud is what we need
        if (kdtree.nearestKSearch(s_point, K, indices, distances) > 0)
        {
            dir = cloud_points->points[i].getVector3fMap() - spline_cloud_proj->points[indices[1]].getVector3fMap();
            dir.normalize();
            directions.push_back(dir);
        }
    }
    return directions;
}


std::vector<Eigen::Vector3f> Pose::computeDirectionsDrum(   pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_points,
                                                            std::vector<Eigen::Vector3f> &center_points,
                                                            pcl::PointCloud<pcl::PointXYZ>::Ptr &spline_cloud_proj)
{
    std::vector<Eigen::Vector3f> directions;
    directions.reserve(cloud_points->size());

    pcl::KdTreeFLANN<pcl::PointXYZ> kdtree;
    kdtree.setInputCloud(spline_cloud_proj);
    int K = 2;
    std::vector<int> indices(K);
    std::vector<float> distances(K);
    pcl::PointXYZ s_point;

    Eigen::Vector3f dir;
    for (size_t i = 0; i < cloud_points->size(); ++i)
    {
        s_point.x = cloud_points->points[i].x;
        s_point.y = cloud_points->points[i].y;
        s_point.z = cloud_points->points[i].z;

        // second point on the spline_cloud is what we need
        if (kdtree.nearestKSearch(s_point, K, indices, distances) > 0)
        {
            // get normal dir and plane
            Eigen::Vector3f normal_dir = computeNormalDir(center_points[i]);
            pcl::ModelCoefficients::Ptr plane(new pcl::ModelCoefficients);         
            plane->values.push_back(normal_dir.x());
            plane->values.push_back(normal_dir.y());
            plane->values.push_back(normal_dir.z());
            plane->values.push_back(-0.1);

            // project temporarly the spline
            pcl::ProjectInliers<pcl::PointXYZ> proj;
            proj.setModelType(pcl::SACMODEL_PLANE);
            proj.setInputCloud(spline_cloud_proj);
            proj.setModelCoefficients(plane);

            pcl::PointCloud<pcl::PointXYZ>::Ptr spline_cloud_tmp(new pcl::PointCloud<pcl::PointXYZ>);
            proj.filter(*spline_cloud_tmp);
            pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_points_tmp(new pcl::PointCloud<pcl::PointXYZ>);
            proj.filter(*cloud_points_tmp);

            dir = cloud_points_tmp->points[i].getVector3fMap() - spline_cloud_tmp->points[indices[1]].getVector3fMap();
            dir.normalize();
            directions.push_back(dir);
        }
    }
    return directions;
}


Eigen::Vector3f Pose::computeNormalDir(Eigen::Vector3f &frame_origin)
{
    Eigen::Vector3f line_pt = {0, 0, 0};
    Eigen::Vector3f line_dir = {1, 0, 0};
    
    //project point onto the line
    Eigen::Vector3f frame_origin_proj = line_pt + (frame_origin - line_pt).dot(line_dir) * line_dir;

    Eigen::Vector3f normal = frame_origin - frame_origin_proj;
    normal.normalize();
    return normal;
}


void Pose::orderOutputMatirces(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_entropy, int task_id,
                                Affine3dVector &transformation_matrix_vector)
{   
    // order output matrices based on entropy value!
    Affine3dVector matrices_ordered;

    std::vector<float> points;
    if (task_id == 1) // BIN
    {
        for(int i = 0; i < transformation_matrix_vector.size(); i++)
            points.push_back(transformation_matrix_vector[i].translation().z());
    
        // sort vector
        std::sort(points.begin(), points.end()); 
        std::reverse(points.begin(), points.end());
    }
    else if (task_id == 2) // WM
    {
        for(int i = 0; i < transformation_matrix_vector.size(); i++)
            points.push_back(transformation_matrix_vector[i].translation().y());

        // sort vector
        std::sort(points.begin(), points.end()); 
        //std::reverse(points.begin(), points.end());
    }
    else if (task_id == 3) // DRUM
    {
        for(int i = 0; i < transformation_matrix_vector.size(); i++)
            points.push_back(transformation_matrix_vector[i].translation().x());
        
        // sort vector
        std::sort(points.begin(), points.end()); 
        std::reverse(points.begin(), points.end());
    }
    else
        std::cout << "task_id not know!" << std::endl;

    std::vector<float> points_copy = points;
    for(int i = 0; i < points.size(); i++){
        std::vector<float>::iterator it = std::find(points_copy.begin(), points_copy.end(), points[i]);
        if (it != points_copy.end()){
            int d = std::distance(points_copy.begin(), it);
            std::cout << d << " " << points_copy[d] << std::endl;

            
            matrices_ordered.push_back(transformation_matrix_vector[d]);
        }
        else 
            std::cout << "not found" << std::endl;
    }
    std::cout << matrices_ordered.size() << std::endl;

    transformation_matrix_vector = matrices_ordered;   
}


void Pose::visualize(bool flag_spin)
{
    pcl::visualization::PCLVisualizer viz("Final_Grasping_Poses");
    viz.setBackgroundColor(1.0f, 1.0f, 1.0f);

    //viz.addPointCloud<pcl::PointXYZRGB>(source_, "source_");
    //viz.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_COLOR, 0.2, 0, 1, "source_");
    //viz.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_OPACITY, 0.8, "source_");

    //pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity(cloud_seg_, "intensity");
    //viz.addPointCloud<pcl::PointXYZI>(cloud_seg_, intensity, "cloud_seg_");
    viz.addPointCloud<pcl::PointXYZI>(cloud_seg_, "cloud_seg_");
    viz.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 10, "cloud_seg_");
    viz.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_COLOR, 0.8f, 0.0f, 0.0f, "cloud_seg_");

    for (int k = 0; k < frames_.size(); k++)
    {
        //viz.addArrow(frames_[k].x, frames_[k].o, 1.0f, 0.0f, 0.0f, false, "x_axis" + std::to_string(k));
        viz.addArrow(frames_[k].y, frames_[k].o, 0.0f, 1.0f, 0.0f, false, "y_axis" + std::to_string(k));
        viz.addArrow(frames_[k].z, frames_[k].o, 0.0f, 0.0f, 1.0f, false, "z_axis" + std::to_string(k));
    }

    for (int k = 0; k < splines_.size(); k++)
    {
        viz.addPointCloud<pcl::PointXYZ>(splines_[k], "splines_" + std::to_string(k));
        viz.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 15, "splines_" + std::to_string(k));
        viz.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_COLOR, 1.0, 0.8, 0.0, "splines_" + std::to_string(k));
        //for (int i = 1; i < splines_[k]->size(); i++)
        //viz.addLine(splines_[k]->points[i - 1], splines_[k]->points[i], 0.56, 0, 1, "BCC" + std::to_string(i) + std::to_string(k));
    }

    if (flag_spin)
        viz.spin();
}

Eigen::Affine3d Pose::computeTransformation(Eigen::Vector3f &origin, Eigen::Vector3f &directionX,
                                            Eigen::Vector3f &directionY, Eigen::Vector3f &directionZ)
{
    directionX.normalize();
    directionZ.normalize();

    Eigen::VectorXd from_line_x, from_line_z, to_line_x, to_line_z;
    from_line_x.resize(6);
    from_line_z.resize(6);
    to_line_x.resize(6);
    to_line_z.resize(6);

    //Origin
    from_line_x << 0, 0, 0, 1, 0, 0;
    from_line_z << 0, 0, 0, 0, 0, 1;

    to_line_x.head<3>() = origin.cast<double>();
    to_line_x.tail<3>() = directionX.cast<double>();

    to_line_z.head<3>() = origin.cast<double>();
    to_line_z.tail<3>() = directionZ.cast<double>();

    Eigen::Affine3d transformation;
    if (!pcl::transformBetween2CoordinateSystems(from_line_x, from_line_z, to_line_x, to_line_z, transformation))
    {
        PCL_WARN("Transformation not found!\n");
    }
    return transformation;
}

FrameT Pose::createFrameForVisualization(Eigen::Vector3f &w_center, Eigen::Vector3f &w_dir, Eigen::Vector3f &w_dir2, Eigen::Vector3f &w_dir3)
{
    float length = 0.025;
    FrameT frame;
    frame.o = {w_center.x(), w_center.y(), w_center.z()};

    frame.x = {w_center.x() + length * w_dir.x(),
               w_center.y() + length * w_dir.y(),
               w_center.z() + length * w_dir.z()};

    frame.y = {w_center.x() + length * w_dir2.x(),
               w_center.y() + length * w_dir2.y(),
               w_center.z() + length * w_dir2.z()};

    frame.z = {w_center.x() + length * w_dir3.x(),
               w_center.y() + length * w_dir3.y(),
               w_center.z() + length * w_dir3.z()};

    return frame;
}
