#include "../include/supervoxel.h"

float SuperVoxel::computeCurvatureScore(std::vector<GraphT> &nodes, size_t last_id, size_t parent_id, size_t candidate_id)
{
    Eigen::Vector3f dir_path = (nodes[parent_id].node.getVector3fMap() - nodes[last_id].node.getVector3fMap());
    Eigen::Vector3f dir_candidate = (nodes[candidate_id].node.getVector3fMap() - nodes[last_id].node.getVector3fMap());
    dir_path.normalize();
    dir_candidate.normalize();

    float angle = std::atan2(dir_path.cross(dir_candidate).norm(), dir_path.dot(dir_candidate));
    return angle;
}

float SuperVoxel::vonmisesPdf(float angle)
{
    return vmises_.pdf(angle);
}

Eigen::Vector3f SuperVoxel::computeDirection(pcl::PointCloud<pcl::PointXYZRGBA>::Ptr cloud_voxel, Eigen::Vector3f centroid)
{
    Eigen::Vector3f direction;

    Eigen::Vector4f centroid4f = {centroid.x(), centroid.y(), centroid.z(), 1};
    Eigen::Matrix3f covariance;
    pcl::computeCovarianceMatrixNormalized(*cloud_voxel, centroid4f, covariance);

    Eigen::Vector3f eigenValues;
    Eigen::Matrix3f eigenVectors;
    pcl::eigen33(covariance, eigenVectors, eigenValues);

    std::vector<double> v;
    v.reserve(3);
    v.emplace_back(eigenValues[0]);
    v.emplace_back(eigenValues[1]);
    v.emplace_back(eigenValues[2]);

    direction = eigenVectors.col(std::max_element(v.begin(), v.end()) - v.begin());
    direction.normalize();

    return direction;
}

pcl::PointCloud<pcl::PointXYZRGBA>::Ptr SuperVoxel::cloudVoxel(size_t target_id,
                                                               std::map<std::uint32_t, pcl::Supervoxel<PointT2>::Ptr> supervoxel_clusters,
                                                               std::multimap<std::uint32_t, std::uint32_t> supervoxel_adjacency)
{
    pcl::PointCloud<pcl::PointXYZRGBA>::Ptr cloud_voxel(new pcl::PointCloud<pcl::PointXYZRGBA>);
    size_t id = 0;
    for (auto label_itr = supervoxel_adjacency.cbegin(); label_itr != supervoxel_adjacency.cend();)
    {
        if (id == target_id)
        {
            std::uint32_t supervoxel_label = label_itr->first;
            pcl::Supervoxel<PointT2>::Ptr supervoxel = supervoxel_clusters.at(supervoxel_label);

            cloud_voxel = supervoxel->voxels_;
            break;
        }
        id++;
    }
    return cloud_voxel;
}

size_t SuperVoxel::nodeMaxIntensitySet(std::vector<GraphT> &set, std::vector<GraphT> &nodes)
{
    //search node with max intensity in the set
    size_t index = 0;
    float max_int = 0;
    for (size_t i = 0; i < set.size(); ++i)
        if (nodes[set[i].id].node.intensity > max_int)
        {
            max_int = nodes[set[i].id].node.intensity;
            index = set[i].id;
        }
    std::cout << "node with max intensity in the set, id: " << index << std::endl;
    return index;
}

std::vector<float> SuperVoxel::neighborsScores(std::vector<GraphT> &set, std::vector<GraphT> &nodes,
                                               pcl::PointCloud<pcl::PointXYZRGBA>::Ptr &cloud_voxel,
                                               size_t index)

{
    // compute direction based on first node
    Eigen::Vector3f direction = computeDirection(cloud_voxel, nodes[index].node.getVector3fMap());

    // use the direction to select next node
    size_t nn_size = nodes[index].neighor_ids.size();
    std::vector<float> curvature_pdfs;
    std::vector<float> intensity_pdfs;
    curvature_pdfs.reserve(nn_size);
    intensity_pdfs.reserve(nn_size);
    for (size_t i = 0; i < nn_size; ++i)
    {
        size_t candidate_id = nodes[index].neighor_ids[i];
        Eigen::Vector3f dir_candidate = (nodes[candidate_id].node.getVector3fMap() - nodes[index].node.getVector3fMap());
        dir_candidate.normalize();

        float angle = std::atan2(direction.cross(dir_candidate).norm(), direction.dot(dir_candidate));
        std::cout << "ANGLE: " << angle << std::endl;
        if (angle < M_PI_2) // to constraint btw M_PI_2 and M_PI
            angle = M_PI - angle;

        float angle_pdf = vonmisesPdf(angle);
        std::cout << "candidate_id: " << candidate_id << ", angle: " << angle << ", angle_pdf: " << angle_pdf << std::endl;

        curvature_pdfs.emplace_back(angle_pdf);
        intensity_pdfs.emplace_back(nodes[candidate_id].node.intensity);
    }

    std::vector<float> product_pdfs;
    product_pdfs.reserve(nn_size);
    float curvature_pdf_threshold = 0.08; // around 65deg
    for (size_t i = 0; i < curvature_pdfs.size(); ++i)
    {
        if (curvature_pdfs[i] < curvature_pdf_threshold)
            product_pdfs.emplace_back(0);
        else
            product_pdfs.emplace_back(curvature_pdfs[i] * intensity_pdfs[i]);
    }

    return product_pdfs;
}

int SuperVoxel::extractTwoNN(std::vector<float> &pdfs, std::vector<GraphT> &nodes, size_t &idx1, size_t &idx2)
{
    std::vector<size_t> out;

    // first nn
    idx1 = std::max_element(pdfs.begin(), pdfs.end()) - pdfs.begin();
    if (pdfs[idx1] == 0)
        return -2;

    // second nn
    pdfs[idx1] = 0;
    idx2 = std::max_element(pdfs.begin(), pdfs.end()) - pdfs.begin();

    if (pdfs[idx2] != 0)
        return 0;
    else
        return -1;
}

size_t SuperVoxel::nodeMinNN(std::vector<GraphT> &nodes)
{
    // search node with lowest number of neighboors
    size_t index = 0;
    int min_nn = 1000;
    for (size_t i = 0; i < nodes.size(); ++i)
    {
        if (nodes[i].neighbor_nodes.size() < min_nn)
        {
            min_nn = nodes[i].neighbor_nodes.size();
            index = i;
        }
    }

    // if min number of neighboors is not 1, look for the node with highest intensity!
    if (nodes[index].neighor_ids.size() != 1)
    {
        float max_int = 0;
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            if (nodes[i].node.intensity > max_int)
            {
                max_int = nodes[i].node.intensity;
                index = i;
            }
        }
    }

    std::cout << "index " << index << std::endl;
    return index;
}

size_t SuperVoxel::nnMaxIntensity(std::vector<size_t> &nn, std::vector<GraphT> &nodes)
{
    // search node with highest intensity among nearest neighbors
    size_t index = 0;
    float max_int = 0;
    for (size_t i = 0; i < nn.size(); ++i)
    {
        if (nodes[nn[i]].node.intensity > max_int)
        {
            max_int = nodes[nn[i]].node.intensity;
            index = nn[i];
        }
    }
    return index;
}

pcl::PointCloud<pcl::PointXYZRGBA>::Ptr SuperVoxel::projectCloud(pcl::PointCloud<pcl::PointXYZRGBA>::Ptr &cloud_in,
                                                                 pcl::ModelCoefficients::Ptr &plane)
{
    pcl::PointCloud<pcl::PointXYZRGBA>::Ptr cloud_proj(new pcl::PointCloud<pcl::PointXYZRGBA>);
    pcl::ProjectInliers<pcl::PointXYZRGBA> proj;
    proj.setModelType(pcl::SACMODEL_PLANE);
    proj.setInputCloud(cloud_);
    proj.setModelCoefficients(plane_);
    proj.filter(*cloud_proj);

    return cloud_proj;
}

pcl::PointCloud<pcl::PointXYZI>::Ptr SuperVoxel::translatePathToCloud(pcl::PointCloud<pcl::PointXYZRGBA>::Ptr &cloud,
                                                                      pcl::PointCloud<pcl::PointXYZRGBA>::Ptr &cloud_proj,
                                                                      pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud_entropy,
                                                                      std::vector<GraphT> &nodes,
                                                                      std::vector<size_t> &path)
{
    pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_output(new pcl::PointCloud<pcl::PointXYZI>);

    pcl::PointXYZI point_out;
    pcl::PointXYZRGBA P;
    pcl::KdTreeFLANN<pcl::PointXYZRGBA> kdtree;
    kdtree.setInputCloud(cloud_proj);
    float radius = 0.01;

    std::vector<int> indices;
    std::vector<float> sqrDistances;
    for (size_t p : path)
    {
        P.x = nodes[p].node.x;
        P.y = nodes[p].node.y;
        P.z = nodes[p].node.z;

        if (kdtree.radiusSearch(P, radius, indices, sqrDistances) > 0)
        {
            float max_int_entropy = cloud_entropy->points[indices[0]].intensity;
            int max_int_entropy_index = 0;
            for (size_t k = 1; k < indices.size(); ++k)
            {
                if (cloud_entropy->points[indices[k]].intensity > max_int_entropy)
                {
                    max_int_entropy = cloud_entropy->points[indices[k]].intensity;
                    max_int_entropy_index = k;
                }
            }
            point_out.x = cloud->points[indices[max_int_entropy_index]].x;
            point_out.y = cloud->points[indices[max_int_entropy_index]].y;
            point_out.z = cloud->points[indices[max_int_entropy_index]].z;
            point_out.intensity = max_int_entropy;
            cloud_output->points.push_back(point_out);
        }
        indices.clear();
    }
    cloud_output->height = 1;
    cloud_output->width = cloud_output->points.size();

    return cloud_output;
}

std::vector<size_t> SuperVoxel::buildPath(std::vector<GraphT> &set, std::vector<GraphT> &nodes, std::vector<size_t> &path)
{
    //float curvature_threshold = M_PI_2 - 0.3;

    // FORWARD LOOP
    bool flag_loop = true;
    while (flag_loop)
    {
        size_t psize = path.size();
        size_t target_id = path[psize - 1];
        std::vector<size_t> target_nn = nodes[target_id].neighor_ids;
        size_t target_nn_size = target_nn.size();

        std::vector<size_t> neighbors;
        for (size_t i = 0; i < target_nn_size; ++i)
        {
            bool flag_already_inside = false;
            for (size_t ii = 0; ii < path.size(); ++ii)
            {
                if (target_nn[i] == path[ii]) // they have the same ID, so it is already inside the path!
                    flag_already_inside = true;
            }
            if (!flag_already_inside)
                neighbors.push_back(target_nn[i]);
        }

        size_t neighbors_size = neighbors.size();
        if (neighbors_size > 0)
        {
            // intensity and curvature (von mises pdf)
            std::vector<float> curvature_pdfs;
            curvature_pdfs.reserve(neighbors_size);

            std::vector<float> product_pdfs;
            product_pdfs.reserve(neighbors_size);

            float curvature_pdf_threshold = 0.08; // around 65deg
            float curvature_score, tot_score;
            for (size_t i = 0; i < neighbors_size; ++i)
            {
                //curvature
                curvature_score = vonmisesPdf(computeCurvatureScore(nodes, path[psize - 1], path[psize - 2], neighbors[i]));
                curvature_pdfs.emplace_back(curvature_score);

                //product
                if (curvature_score < curvature_pdf_threshold)
                    tot_score = 0;
                else
                    tot_score = curvature_score * nodes[neighbors[i]].node.intensity;
                product_pdfs.emplace_back(tot_score);
                //std::cout << tot_score << " " << curvature_score << " " << nodes[neighbors[i]].node.intensity << std::endl;
            }

            size_t idx_max = std::max_element(product_pdfs.begin(), product_pdfs.end()) - product_pdfs.begin();
            if (product_pdfs[idx_max] > 0)
            {
                //std::cout << "node id selected: " << neighbors[idx_max] << std::endl;
                path.push_back(neighbors[idx_max]);
            }
            else
            {
                flag_loop = false;
                std::cout << "product PDFs is zero for all possible neighbors!" << std::endl;
            }
        }
        else
        {
            flag_loop = false;
            std::cout << "not neighbors found! " << std::endl;
        }

        /*

        size_t neighbors_size = neighbors.size();
        // FLOOD-FILL if only one neighbor left
        if (neighbors_size == 1)
        {
            //curvature
            float curvature_score = vonmisesPdf(computeCurvatureScore(nodes, path[psize - 1], path[psize - 2], neighbors[0]));

            //product
            float curvature_pdf_threshold = 0.15;
            if (curvature_score < curvature_pdf_threshold)
                flag_loop = false;
            else
                path.push_back(neighbors[0]);
        }

        // we have multiple neighbors to decide with!
        else if (neighbors_size > 1)
        {
            // intensity and curvature (von mises pdf)
            std::vector<float> curvature_pdfs;
            curvature_pdfs.reserve(neighbors_size);

            std::vector<float> product_pdfs;
            product_pdfs.reserve(neighbors_size);

            float curvature_pdf_threshold = 0.08; // around 65deg
            float curvature_score, tot_score;
            for (size_t i = 0; i < neighbors_size; ++i)
            {
                //curvature
                curvature_score = vonmisesPdf(computeCurvatureScore(nodes, path[psize - 1], path[psize - 2], neighbors[i]));
                curvature_pdfs.emplace_back(curvature_score);

                //product
                if (curvature_score < curvature_pdf_threshold)
                    tot_score = 0;
                else
                    tot_score = curvature_score * nodes[neighbors[i]].node.intensity;
                product_pdfs.emplace_back(tot_score);

                }

                size_t idx_max = std::max_element(product_pdfs.begin(), product_pdfs.end()) - product_pdfs.begin();
                if (product_pdfs[idx_max] > 0)
                {
                    //std::cout << "node id selected: " << neighbors[idx_max] << std::endl;
                    path.push_back(neighbors[idx_max]);
                }
                else
                {
                    flag_loop = false;
                    std::cout << "product PDFs is zero for all possible neighbors!" << std::endl;
                }
            }
            */
    }

    std::cout << "FORWARD_PATH: ";
    for (auto p : path)
        std::cout << p << " ";
    std::cout << std::endl;

    return path;
}

bool SuperVoxel::computeSupervoxels(std::vector<pcl::PointCloud<pcl::PointXYZI>::Ptr> &clouds_output)
{

    // set cloud
    cloud_ = colorCloud(cloud_entropy_);
   
    if (task_id_ == 1 || task_id_ == 2) // bin or washing machine
        cloud_proj_ = projectCloud(cloud_, plane_);
    else if (task_id_ == 3) // drum
        cloud_proj_ = cloud_;  
    else
        PCL_WARN("Uknown task_id assigned!");

    // SuperVoxels ---------------------------------------------------------------
    pcl::SupervoxelClustering<PointT2> super(voxel_resolution_, seed_resolution_);
    if (disable_transform_)
        super.setUseSingleCameraTransform(false);
    super.setInputCloud(cloud_proj_);
    super.setColorImportance(color_importance_);
    super.setSpatialImportance(spatial_importance_);
    super.setNormalImportance(normal_importance_);

    std::map<std::uint32_t, pcl::Supervoxel<PointT2>::Ptr> supervoxel_clusters;

    pcl::console::print_highlight("Extracting supervoxels!\n");
    super.extract(supervoxel_clusters);
    pcl::console::print_info("Found %d supervoxels\n", supervoxel_clusters.size());
    pcl::console::print_highlight("Getting supervoxel adjacency\n");
    std::multimap<std::uint32_t, std::uint32_t> supervoxel_adjacency;
    super.getSupervoxelAdjacency(supervoxel_adjacency);

    PointCloudT2::Ptr voxel_centroid_cloud = super.getVoxelCentroidCloud();
    PointLCloudT::Ptr labeled_voxel_cloud = super.getLabeledVoxelCloud();

    /*
    pcl::visualization::PCLVisualizer::Ptr viewer(new pcl::visualization::PCLVisualizer("3D Viewer"));
    viewer->setBackgroundColor(0, 0, 0);
    viewer->addPointCloud(voxel_centroid_cloud, "voxel centroids");
    viewer->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 2.0, "voxel centroids");
    viewer->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_OPACITY, 0.95, "voxel centroids");
    viewer->addPointCloud(labeled_voxel_cloud, "labeled voxels");
    viewer->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_OPACITY, 0.8, "labeled voxels");
    */
   
    // Building Graph --------------------------------------------------------------
    std::vector<GraphT> nodes = populateGraph(supervoxel_clusters, supervoxel_adjacency);

    nodes_ = nodes;
    displayNodesInformation(nodes);
    cloud_nodes_ = getNodesCloud(nodes); // only for visualization

    std::vector<std::vector<GraphT>> sets;
    splitSetNodes(nodes, sets);



    for (auto set : sets)
    {
        // FIRST NODE
        size_t index = nodeMaxIntensitySet(set, nodes);

        // get cloud of that node
        pcl::PointCloud<pcl::PointXYZRGBA>::Ptr cloud_voxel = cloudVoxel(index, supervoxel_clusters, supervoxel_adjacency);

        // SECOND NODE
        std::vector<float> nn_index_scores = neighborsScores(set, nodes, cloud_voxel, index);
        for(int i = 0; i < nn_index_scores.size(); i++)
            std::cout << nn_index_scores[i] << std::endl;

        size_t idx1, idx2;
        int result = extractTwoNN(nn_index_scores, nodes, idx1, idx2);

        if (result == 0) // both idx1 and idx2 are valid
        {
            std::cout << "result == 0" << std::endl;
            std::vector<size_t> path1, path2;
            path1.push_back(index);                          // first node
            path1.push_back(nodes[index].neighor_ids[idx1]); // second node

            path2.push_back(index);                          // first node
            path2.push_back(nodes[index].neighor_ids[idx2]); // second node

            // BUILD PATH ------------------------------------------------------------------
            buildPath(set, nodes, path1);
            buildPath(set, nodes, path2);

            // MERGING the two paths
            bool flag_stop = false;
            if (path1[0] == path2[0]) // they have the same node
            {
                for (size_t k = 1; k < path2.size(); ++k) // k=1 since we want to skip the same id
                {
                    flag_stop = false;
                    for (auto p : path1)
                        if (p == path2[k])
                            flag_stop = true;

                    if (!flag_stop)
                        path1.insert(path1.begin(), path2[k]);
                }
            }
            std::cout << "FINAL_PATH: ";
            for (auto p : path1)
                std::cout << p << " ";
            std::cout << std::endl;

            // OUTPUT CLOUD
            pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_output = translatePathToCloud(cloud_, cloud_proj_, cloud_entropy_, nodes, path1);
            clouds_output.push_back(cloud_output);
        }
        else if (result == -1) // only idx1 is valid
        {
            std::cout << "result == -1" << std::endl;
            std::vector<size_t> path;
            path.push_back(index);                          // first node
            path.push_back(nodes[index].neighor_ids[idx1]); // second node

            // BUILD PATH ------------------------------------------------------------------
            buildPath(set, nodes, path);

            // OUTPUT CLOUD
            pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_output = translatePathToCloud(cloud_, cloud_proj_, cloud_entropy_, nodes, path);
            clouds_output.push_back(cloud_output);
        }
        else if (result == -2) 
            continue;
    }
    if (clouds_output.size() != 0)
        return true;
    else
        return false;
    
}

void SuperVoxel::updateNodesIDs(std::vector<GraphT> &nodes)
{
    Eigen::Vector3f vec1, vec2;
    for (size_t i = 0; i < nodes.size(); ++i)
    {
        // loop throght neighbors
        for (size_t j = 0; j < nodes.size(); ++j)
        {
            vec2 = nodes[j].node.getVector3fMap();
            for (size_t k = 0; k < nodes[i].neighbor_nodes.size(); ++k)
            {
                vec1 = nodes[i].neighbor_nodes[k].getVector3fMap();
                if (vec1.isApprox(vec2))
                    nodes[i].neighor_ids[k] = nodes[j].id;
            }
        }
    }
}

pcl::PointCloud<pcl::PointXYZI>::Ptr SuperVoxel::getNodesCloud(std::vector<GraphT> nodes)
{
    pcl::PointCloud<pcl::PointXYZI>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZI>);
    for (size_t it = 0; it < nodes.size(); ++it)
    {
        cloud->points.push_back(nodes[it].node);
    }
    return cloud;
}

// extract supervoxel's centroid and neighbors
std::vector<GraphT> SuperVoxel::populateGraph(std::map<std::uint32_t, pcl::Supervoxel<PointT2>::Ptr> supervoxel_clusters,
                                              std::multimap<std::uint32_t, std::uint32_t> supervoxel_adjacency)
{
    size_t id = 0;
    GraphT node_temp;
    std::vector<GraphT> nodes;
    for (auto label_itr = supervoxel_adjacency.cbegin(); label_itr != supervoxel_adjacency.cend();)
    {
        std::uint32_t supervoxel_label = label_itr->first;
        pcl::Supervoxel<PointT2>::Ptr supervoxel = supervoxel_clusters.at(supervoxel_label);

        node_temp.node = nnSearch(supervoxel);
        node_temp.id = id;

        for (auto adjacent_itr = supervoxel_adjacency.equal_range(supervoxel_label).first; adjacent_itr != supervoxel_adjacency.equal_range(supervoxel_label).second; ++adjacent_itr)
        {
            pcl::Supervoxel<PointT2>::Ptr neighbor_supervoxel = supervoxel_clusters.at(adjacent_itr->second);
            node_temp.neighbor_nodes.push_back(nnSearch(neighbor_supervoxel));
            node_temp.neighor_ids.push_back(1000); // huge value since it will be assigned later
        }

        label_itr = supervoxel_adjacency.upper_bound(supervoxel_label);

        nodes.push_back(node_temp);
        node_temp.neighbor_nodes.clear();
        node_temp.neighor_ids.clear();
        id++;
    }

    // finalize node ID assignment
    updateNodesIDs(nodes);

    // normalize intensity values between 0 - 1
    //normalizeIntensity(nodes);

    return nodes;
}

// return pcl::PointXYZI where XYZ are the supervoxel centroid and I is the avera points intensities.
// for each points in supervoxel, search the nearest neighbor in the input pcl::PointXYZI cloud
// output intensity is the maximum value found in the supervoxel "nearest neighbor" points
pcl::PointXYZI SuperVoxel::nnSearch(pcl::Supervoxel<PointT2>::Ptr supervoxel)
{
    pcl::PointXYZI point_temp;
    pcl::PointXYZRGBA searchPoint;
    float sum_intensities = 0;
    size_t counter = 0;

    pcl::KdTreeFLANN<pcl::PointXYZRGBA> kdtree;
    kdtree.setInputCloud(cloud_proj_);
    int K = 1;
    std::vector<int> indices(K);
    std::vector<float> distances(K);

    //std::vector<float> intensities;
    for (int i = 0; i < supervoxel->voxels_->size(); ++i)
    {
        searchPoint = supervoxel->voxels_->points[i];
        if (kdtree.nearestKSearch(searchPoint, K, indices, distances) > 0)
        {
            //intensities.push_back(cloud_entropy_->points[indices[0]].intensity);
            //if (max_intensity < cloud_entropy_->points[indices[0]].intensity)
            //    max_intensity = cloud_entropy_->points[indices[0]].intensity;
            sum_intensities += cloud_entropy_->points[indices[0]].intensity;
            counter++;
        }
        indices.clear();
        distances.clear();
    }
    //size_t n = intensities.size() / 2;
    //std::nth_element(intensities.begin(), intensities.begin() + n, intensities.end());

    point_temp.getVector3fMap() = supervoxel->centroid_.getVector3fMap();
    point_temp.intensity = sum_intensities / counter;
    return point_temp;
}

void SuperVoxel::normalizeIntensity(std::vector<GraphT> &nodes)
{
    std::vector<float> vec;
    for (size_t i = 0; i < nodes.size(); ++i)
    {
        vec.push_back(nodes[i].node.intensity);
    }

    float max = *std::max_element(vec.begin(), vec.end());
    float min = *std::min_element(vec.begin(), vec.end());

    for (size_t i = 0; i < nodes.size(); ++i)
        nodes[i].node.intensity = (vec[i] - min) / (max - min);
}

void SuperVoxel::displayNodesInformation(std::vector<GraphT> nodes)
{
    for (size_t i = 0; i < nodes.size(); ++i)
    {
        std::cout.precision(3);
        std::cout << "Node ID: " << nodes[i].id << " | intensity: " << nodes[i].node.intensity << "| ";
        for (size_t k = 0; k < nodes[i].neighor_ids.size(); ++k)
            std::cout << nodes[i].neighor_ids[k] << " ";

        std::cout << std::endl;
    }
}

// VISUALIZATION

void SuperVoxel::visualization()
{
    // VISUALIZATION
    pcl::visualization::PCLVisualizer::Ptr viz(new pcl::visualization::PCLVisualizer("Nodes"));
    viz->setBackgroundColor(0.0f, 0.0f, 0.0f);
    pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity_distribution(cloud_nodes_, "intensity");
    viz->addPointCloud<pcl::PointXYZI>(cloud_nodes_, intensity_distribution, "cloud_nodes_");
    viz->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 3, "cloud_nodes_");
    viz->addPointCloud<pcl::PointXYZRGBA>(cloud_, "cloud_");
    viz->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_COLOR, 0.0f, 0.4f, 0.0f, "cloud_");

    for (size_t i = 0; i < nodes_.size(); ++i)
    {
        std::string name = std::to_string(nodes_[i].id);
        viz->addText3D(name, nodes_[i].node, 0.01);
    }
    for (size_t i = 1; i < sequence_id_viz_.size(); ++i)
        viz->addLine(nodes_[sequence_id_viz_[i - 1]].node, nodes_[sequence_id_viz_[i]].node, "line" + std::to_string(i));

    pcl::visualization::PCLVisualizer::Ptr vizLines(new pcl::visualization::PCLVisualizer("All Lines"));
    vizLines->setBackgroundColor(1.0f, 1.0f, 1.0f);
    pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity_distributionLines(cloud_nodes_, "intensity");
    vizLines->addPointCloud<pcl::PointXYZI>(cloud_nodes_, intensity_distributionLines, "cloud_nodes_");
    vizLines->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 10, "cloud_nodes_");
    vizLines->addPointCloud<pcl::PointXYZRGBA>(cloud_, "cloud_");
    vizLines->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_COLOR, 0.0f, 0.4f, 0.0f, "cloud_");
    //vizLines->addPlane(*plane_, "plane");

    drawAllLines(vizLines);

    pcl::visualization::PCLVisualizer::Ptr vizLinesP(new pcl::visualization::PCLVisualizer("All Lines Proj"));
    vizLinesP->setBackgroundColor(1.0f, 1.0f, 1.0f);
    pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity_distributionLinesP(cloud_nodes_, "intensity");
    vizLinesP->addPointCloud<pcl::PointXYZI>(cloud_nodes_, intensity_distributionLinesP, "cloud_nodes_");
    vizLinesP->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 15, "cloud_nodes_");
    vizLinesP->addPointCloud<pcl::PointXYZRGBA>(cloud_proj_, "cloud_proj_");
    vizLinesP->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_COLOR, 0.8f, 0.0f, 0.0f, "cloud_proj_");
    vizLinesP->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 10, "cloud_proj_");
    //vizLines->addPlane(*plane_, "plane");
    drawAllLines(vizLinesP);

    /*
    pcl::visualization::PCLVisualizer::Ptr vizProj(new pcl::visualization::PCLVisualizer("Proj"));
    vizProj->setBackgroundColor(0.0f, 0.0f, 0.0f);
    pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity_distributionP(cloud_proj, "intensity");
    vizProj->addPointCloud<pcl::PointXYZI>(cloud_proj, intensity_distributionP, "proj");
    vizProj->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 10, "proj");
    */

    /*
    while (!viz->wasStopped())
    {
        viz->spinOnce(100);
    }
    */
}

void SuperVoxel::drawAllLines(pcl::visualization::PCLVisualizer::Ptr viz)
{
    for (int i = 0; i < nodes_.size(); ++i)
        for (int j = 0; j < nodes_[i].neighbor_nodes.size(); ++j)
            viz->addLine(nodes_[i].node, nodes_[i].neighbor_nodes[j], "line" + std::to_string(i) + std::to_string(j));
}

void SuperVoxel::splitSetNodes(std::vector<GraphT> &nodes, std::vector<std::vector<GraphT>> &sets)
{
    std::vector<GraphT> set_tmp;

    bool not_all_assigned = true;
    size_t counter_while = 0;
    while (not_all_assigned)
    {
        // select seed point for start assignent
        for (size_t n = 0; n < nodes.size(); ++n)
        {
            if (nodes[n].assigned == false)
            {
                set_tmp.push_back(nodes[n]);
                nodes[n].assigned = true;
                break;
            }
        }

        // loop to assign nodes
        bool already_inside = false;
        for (size_t j = 0; j < 100; j++)
        {
            for (size_t ii = 0; ii < set_tmp.size(); ++ii)
            {
                GraphT candidate = set_tmp[ii];
                for (size_t i = 0; i < candidate.neighor_ids.size(); ++i)
                {
                    already_inside = false;
                    for (size_t k = 0; k < set_tmp.size(); ++k)
                        if (candidate.neighor_ids[i] == set_tmp[k].id)
                        {
                            already_inside = true;
                            break;
                        }

                    if (!already_inside)
                    {
                        set_tmp.push_back(nodes[candidate.neighor_ids[i]]);
                        nodes[candidate.neighor_ids[i]].assigned = true;
                    }
                }
            }
        }
        // push back set found
        sets.push_back(set_tmp);

        // print set found

        std::cout << "set " << counter_while << ": ";
        for (GraphT s : set_tmp)
            std::cout << s.id << " ";
        std::cout << std::endl;
        counter_while++;

        // clear vector
        set_tmp.clear();

        // check not_all_assigned condition
        size_t tot_size = 0;
        for (auto &v : sets)
            tot_size += v.size();

        if (tot_size == nodes.size())
            not_all_assigned = false;
    }

    std::cout << "nodes: ";
    for (GraphT g : nodes)
        std::cout << g.id << " ";
    std::cout << std::endl;
}




ColourT SuperVoxel::getColour(double v, double vmin, double vmax)
{
    ColourT c = {1.0, 1.0, 1.0}; // white
    double dv;

    if (v < vmin)
        v = vmin;
    if (v > vmax)
        v = vmax;
    dv = vmax - vmin;

    if (v < (vmin + 0.25 * dv))
    {
        c.r = 0;
        c.g = 4 * (v - vmin) / dv;
    }
    else if (v < (vmin + 0.5 * dv))
    {
        c.r = 0;
        c.b = 1 + 4 * (vmin + 0.25 * dv - v) / dv;
    }
    else if (v < (vmin + 0.75 * dv))
    {
        c.r = 4 * (v - vmin - 0.5 * dv) / dv;
        c.b = 0;
    }
    else
    {
        c.g = 1 + 4 * (vmin + 0.75 * dv - v) / dv;
        c.b = 0;
    }

    return c;
}

pcl::PointCloud<pcl::PointXYZRGBA>::Ptr SuperVoxel::colorCloud(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud)
{
    pcl::PointCloud<pcl::PointXYZRGBA>::Ptr cloud_rgb(new pcl::PointCloud<pcl::PointXYZRGBA>);
    cloud_rgb->height = 1;
    cloud_rgb->width = cloud->size();
    cloud_rgb->resize(cloud_rgb->height * cloud_rgb->width);

    pcl::PointXYZRGBA p;
    ColourT col;
    for (std::size_t i = 0; i < cloud->size(); ++i)
    {
        col = getColour(cloud->points[i].intensity, 0, 1);

        p.x = cloud->points[i].x;
        p.y = cloud->points[i].y;
        p.z = cloud->points[i].z;

        p.r = static_cast<uint8_t>(col.b * 255);
        p.g = static_cast<uint8_t>(col.g * 255);
        p.b = static_cast<uint8_t>(col.r * 255);

        p.a = cloud->points[i].intensity;

        cloud_rgb->points[i] = p;
    }
    return cloud_rgb;
}
